# -*- coding: utf-8 -*-
"""This module contains plotting functions based on matplotlib
for image, line, and scatter plots.

A feature of these plotting functions is that
they can be controlled using only hotkeys
so the user does not need to move away from the keyboard.

Given an array ``x``, an example usage is:

    >>> ImagePlot(x)
    >>> LinePlot(x)
    >>> ScatterPlot(x)

"""
import datetime
import os
import subprocess
import uuid

import numpy as np

import sigpy as sp

__all__ = ["ImagePlot", "LinePlot", "ScatterPlot"]


image_plot_help_str = r"""
$\bf{Hotkeys:}$
    $\bf{h:}$ show/hide hotkey menu.
    $\bf{x/y/z:}$ set current axis as x/y/z.
    $\bf{t:}$ swap between x and y.
    $\bf{c:}$ select current axis as color.
    $\bf{left/right:}$ change current axis.
    $\bf{up/down:}$ change slice along current axis.
    $\bf{a:}$ toggle hide all labels, titles and axes.
    $\bf{m/p/r/i/l:}$  magnitude/phase/real/imaginary/log mode.
    $\bf{[/]:}$ change brightness.
    $\bf{\{/\}:}$ change contrast.
    $\bf{s:}$ save as png.
    $\bf{g/v:}$ save as gif/video by along current axis.
    $\bf{q:}$ refresh.
    $\bf{0-9:}$ enter slice number.
    $\bf{enter:}$ set current axis as slice number.
"""


class ImagePlot(object):
    """Plot array as image.

    Press 'h' for a menu for hotkeys.

    Args:
        im (array): image numpy/cupy array.
        x (int): x axis.
        y (int): y axis.
        z (None or int): z axis.
        c (None or int): color axis.
        hide_axes (bool): toggle hiding axes, labels and title.
        mode (str): specify magnitude, phase, real, imaginary,
            and log mode. {'m', 'p', 'r', 'i', 'l'}.
        title (str): title.
        interpolation (str): plot interpolation.
        save_basename (str): saved png, gif, and video base name.
        fps (int): frame per seconds for gif and video.

    """

    def __init__(
        self,
        im,
        x=-1,
        y=-2,
        z=None,
        c=None,
        hide_axes=False,
        mode=None,
        colormap=None,
        vmin=None,
        vmax=None,
        title="",
        interpolation="nearest",
        save_basename="Figure",
        fps=10,
    ):
        if im.ndim < 2:
            raise TypeError(
                "Image dimension must at least be two, got {im_ndim}".format(
                    im_ndim=im.ndim
                )
            )
        import matplotlib.pyplot as plt

        self.axim = None
        self.im = im
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.shape = self.im.shape
        self.ndim = self.im.ndim
        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.x = x % self.ndim
        self.y = y % self.ndim
        self.z = z % self.ndim if z is not None else None
        self.c = c % self.ndim if c is not None else None
        self.d = max(self.ndim - 3, 0)
        self.hide_axes = hide_axes
        self.show_help = False
        self.title = title
        self.interpolation = interpolation
        self.mode = mode
        self.colormap = colormap
        self.entering_slice = False
        self.vmin = vmin
        self.vmax = vmax
        self.save_basename = save_basename
        self.fps = fps
        self.help_text = None

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id
        )
        self.fig.canvas.mpl_connect("key_press_event", self.key_press)
        self.update_axes()
        self.update_image()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == "up":
            if self.d not in [self.x, self.y, self.z, self.c]:
                self.slices[self.d] = (self.slices[self.d] + 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "down":
            if self.d not in [self.x, self.y, self.z, self.c]:
                self.slices[self.d] = (self.slices[self.d] - 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "left":
            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "right":
            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "x" and self.d not in [self.x, self.z, self.c]:
            if self.d == self.y:
                self.x, self.y = self.y, self.x
            else:
                self.x = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "y" and self.d not in [self.y, self.z, self.c]:
            if self.d == self.x:
                self.x, self.y = self.y, self.x
            else:
                self.y = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "z" and self.d not in [self.x, self.y, self.c]:
            if self.d == self.z:
                self.z = None
            else:
                self.z = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif (
            event.key == "c"
            and self.d not in [self.x, self.y, self.z]
            and self.shape[self.d] == 3
        ):
            if self.d == self.c:
                self.c = None
            else:
                self.c = self.d

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "t":
            self.x, self.y = self.y, self.x

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "a":
            self.hide_axes = not self.hide_axes

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "f":
            self.fig.canvas.manager.full_screen_toggle()

        elif event.key == "q":
            self.vmin = None
            self.vmax = None
            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "]":
            width = self.vmax - self.vmin
            self.vmin -= width * 0.1
            self.vmax -= width * 0.1

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "[":
            width = self.vmax - self.vmin
            self.vmin += width * 0.1
            self.vmax += width * 0.1

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "}":
            width = self.vmax - self.vmin
            center = (self.vmax + self.vmin) / 2
            self.vmin = center - width * 1.1 / 2
            self.vmax = center + width * 1.1 / 2

            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "{":
            width = self.vmax - self.vmin
            center = (self.vmax + self.vmin) / 2
            self.vmin = center - width * 0.9 / 2
            self.vmax = center + width * 0.9 / 2

            self.update_image()
            self.fig.canvas.draw()

        elif event.key in ["m", "p", "r", "i", "l"]:
            self.vmin = None
            self.vmax = None
            self.mode = event.key

            self.update_axes()
            self.update_image()
            self.fig.canvas.draw()

        elif event.key == "s":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.png"
            )
            self.fig.savefig(
                filename,
                transparent=True,
                format="png",
                bbox_inches="tight",
                pad_inches=0,
            )

        elif event.key == "g":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.gif"
            )
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_image()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    bbox_inches=bbox,
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vf",
                    "palettegen",
                    "{} palette.png".format(temp_basename),
                ]
            )
            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-i",
                    "{} palette.png".format(temp_basename),
                    "-lavfi",
                    "paletteuse",
                    filename,
                ]
            )

            os.remove("{} palette.png".format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))

        elif event.key == "v":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.mp4"
            )
            temp_basename = uuid.uuid4()

            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_image()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    transparent=True,
                    bbox_inches="tight",
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-r",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vf",
                    "crop=floor(iw/2)*2-10:floor(ih/2)*2-10",
                    "-pix_fmt",
                    "yuv420p",
                    "-crf",
                    "1",
                    "-vcodec",
                    "libx264",
                    "-preset",
                    "veryslow",
                    filename,
                ]
            )

            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))

        elif event.key in [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "backspace",
        ] and self.d not in [self.x, self.y, self.z, self.c]:
            if self.entering_slice:
                if event.key == "backspace":
                    if self.entered_slice < 10:
                        self.entering_slice = False
                    else:
                        self.entered_slice //= 10
                else:
                    self.entered_slice = self.entered_slice * 10 + int(
                        event.key
                    )
            elif event.key != "backspace":
                self.entering_slice = True
                self.entered_slice = int(event.key)

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "enter" and self.entering_slice:
            self.entering_slice = False
            if self.entered_slice < self.shape[self.d]:
                self.slices[self.d] = self.entered_slice

                self.update_image()

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "h":
            self.show_help = not self.show_help

            self.update_image()
            self.fig.canvas.draw()
        else:
            return

    def update_image(self):
        # Extract slice.
        idx = []
        for i in range(self.ndim):
            if i in [self.x, self.y, self.z, self.c]:
                idx.append(slice(None, None, self.flips[i]))
            else:
                idx.append(self.slices[i])

        idx = tuple(idx)
        imv = sp.to_device(self.im[idx])

        # Transpose to have [z, y, x, c].
        imv_dims = [self.y, self.x]
        if self.z is not None:
            imv_dims = [self.z] + imv_dims

        if self.c is not None:
            imv_dims = imv_dims + [self.c]

        imv = np.transpose(imv, np.argsort(np.argsort(imv_dims)))
        imv = array_to_image(imv, color=self.c is not None)

        if self.mode is None:
            if np.isrealobj(imv):
                self.mode = "r"
            else:
                self.mode = "m"

        if self.mode == "m":
            imv = np.abs(imv)
        elif self.mode == "p":
            imv = np.angle(imv)
        elif self.mode == "r":
            imv = np.real(imv)
        elif self.mode == "i":
            imv = np.imag(imv)
        elif self.mode == "l":
            imv = np.abs(imv)
            imv = np.log(imv, out=np.ones_like(imv) * -31, where=imv != 0)

        if self.vmin is None:
            self.vmin = imv.min()

        if self.vmax is None:
            self.vmax = imv.max()

        if self.axim is None:
            if self.colormap is None:
                colormap = "gray"
            else:
                colormap = self.colormap
            self.axim = self.ax.imshow(
                imv,
                vmin=self.vmin,
                vmax=self.vmax,
                cmap=colormap,
                origin="lower",
                interpolation=self.interpolation,
                aspect=1.0,
                extent=[0, imv.shape[1], 0, imv.shape[0]],
            )

            if self.colormap is not None:
                self.fig.colorbar(self.axim)

        else:
            self.axim.set_data(imv)
            self.axim.set_extent([0, imv.shape[1], 0, imv.shape[0]])
            self.axim.set_clim(self.vmin, self.vmax)

        if self.help_text is None:
            bbox_props = dict(
                boxstyle="round", pad=1, fc="white", alpha=0.95, lw=0
            )
            self.help_text = self.ax.text(
                imv.shape[0] / 2,
                imv.shape[1] / 2,
                image_plot_help_str,
                ha="center",
                va="center",
                linespacing=1.5,
                ma="left",
                size=8,
                bbox=bbox_props,
            )

        self.help_text.set_visible(self.show_help)

    def update_axes(self):
        if not self.hide_axes:
            caption = "["
            for i in range(self.ndim):
                if i == self.d:
                    caption += "["
                else:
                    caption += " "

                if self.flips[i] == -1 and (
                    i == self.x or i == self.y or i == self.z or i == self.c
                ):
                    caption += "-"

                if i == self.x:
                    caption += "x"
                elif i == self.y:
                    caption += "y"
                elif i == self.z:
                    caption += "z"
                elif i == self.c:
                    caption += "c"
                elif i == self.d and self.entering_slice:
                    caption += str(self.entered_slice) + "_"
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += "]"
                else:
                    caption += " "
            caption += "]"

            self.ax.set_title(caption)
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title("")
            self.fig.suptitle("")
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)


def mosaic_shape(batch):
    mshape = [int(batch**0.5), batch // int(batch**0.5)]

    while sp.prod(mshape) < batch:
        mshape[1] += 1

    if (mshape[0] - 1) * (mshape[1] + 1) == batch:
        mshape[0] -= 1
        mshape[1] += 1

    return tuple(mshape)


def array_to_image(arr, color=False):
    """
    Flattens all dimensions except the last two

    Args:
        arr (array): shape [z, x, y, c] if color, else [z, x, y]

    """
    if color and not (arr.max() == 0 and arr.min() == 0):
        arr = arr / np.abs(arr).max()

    if arr.ndim == 2:
        return arr
    elif color and arr.ndim == 3:
        return arr

    if color:
        img_shape = arr.shape[-3:]
        batch = sp.prod(arr.shape[:-3])
        mshape = mosaic_shape(batch)
    else:
        img_shape = arr.shape[-2:]
        batch = sp.prod(arr.shape[:-2])
        mshape = mosaic_shape(batch)

    if sp.prod(mshape) == batch:
        img = arr.reshape((batch,) + img_shape)
    else:
        img = np.zeros((sp.prod(mshape),) + img_shape, dtype=arr.dtype)
        img[:batch, ...] = arr.reshape((batch,) + img_shape)

    img = img.reshape(mshape + img_shape)
    if color:
        img = np.transpose(img, (0, 2, 1, 3, 4))
        img = img.reshape(
            (img_shape[0] * mshape[0], img_shape[1] * mshape[1], 3)
        )
    else:
        img = np.transpose(img, (0, 2, 1, 3))
        img = img.reshape((img_shape[0] * mshape[0], img_shape[1] * mshape[1]))

    return img


class LinePlot(object):
    """Plot array as lines.

    Keyword Args:
        x: select current dimension as x
        left/right: increment/decrement current dimension
        up/down: flip axis when current dimension is x or y
            otherwise increment/decrement slice at current dimension
        h: toggle hide all labels, titles and axes
        m: magnitude mode
        p: phase mode
        r: real mode
        i: imaginary mode
        l: log mode
        s: save as png.
        g: save as gif by traversing current dimension.
        v: save as video by traversing current dimension.
    """

    def __init__(
        self,
        arr,
        x=-1,
        hide_axes=False,
        mode="m",
        title="",
        save_basename="Figure",
        fps=10,
    ):
        import matplotlib.pyplot as plt

        self.arr = arr
        self.axarr = None

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.shape = self.arr.shape
        self.ndim = self.arr.ndim
        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.x = x % self.ndim
        self.d = max(self.ndim - 3, 0)
        self.hide_axes = hide_axes
        self.title = title
        self.mode = mode
        self.save_basename = save_basename
        self.fps = fps
        self.bottom = None
        self.top = None

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id
        )
        self.fig.canvas.mpl_connect("key_press_event", self.key_press)
        self.update_axes()
        self.update_line()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == "up":
            if self.d != self.x:
                self.slices[self.d] = (self.slices[self.d] + 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == "down":
            if self.d != self.x:
                self.slices[self.d] = (self.slices[self.d] - 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == "left":
            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "right":
            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "x" and self.d != self.x:
            self.x = self.d

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == "a":
            self.hide_axes = not self.hide_axes

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "f":
            self.fig.canvas.manager.full_screen_toggle()

        elif (
            event.key == "m"
            or event.key == "p"
            or event.key == "r"
            or event.key == "i"
            or event.key == "l"
        ):
            self.mode = event.key
            self.bottom = None
            self.top = None

            self.update_axes()
            self.update_line()
            self.fig.canvas.draw()

        elif event.key == "s":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.png"
            )
            self.fig.savefig(
                filename,
                transparent=True,
                format="png",
                bbox_inches="tight",
                pad_inches=0,
            )

        elif event.key == "g":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.gif"
            )
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_line()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    bbox_inches=bbox,
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vf",
                    "palettegen",
                    "{} palette.png".format(temp_basename),
                ]
            )
            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-i",
                    "{} palette.png".format(temp_basename),
                    "-lavfi",
                    "paletteuse",
                    filename,
                ]
            )

            os.remove("{} palette.png".format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))

        elif event.key == "v":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %h.%M.%S %p.mov"
            )
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_line()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    bbox_inches=bbox,
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vcodec",
                    "png",
                    filename,
                ]
            )

            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))
        else:
            return

        return

    def update_line(self):
        order = [i for i in range(self.ndim) if i != self.x] + [self.x]
        idx = tuple(
            [self.slices[i] for i in order[:-1]]
            + [slice(None, None, self.flips[self.x])]
        )

        arrv = self.arr.transpose(order)[idx]

        if self.mode == "m":
            arrv = np.abs(arrv)
        elif self.mode == "p":
            arrv = np.angle(arrv)
        elif self.mode == "r":
            arrv = np.real(arrv)
        elif self.mode == "i":
            arrv = np.imag(arrv)
        elif self.mode == "l":
            eps = 1e-31
            arrv = np.log(np.abs(arrv) + eps)

        if self.bottom is None:
            self.bottom = arrv.min()

        if self.top is None:
            self.top = arrv.max()

        if self.axarr is None:
            self.axarr = self.ax.plot(arrv)[0]

        else:
            self.axarr.set_xdata(np.arange(len(arrv)))
            self.axarr.set_ydata(arrv)
            self.ax.set_ylim(self.bottom, self.top)

    def update_axes(self):
        if not self.hide_axes:
            caption = "Slice: ["
            for i in range(self.ndim):
                if i == self.d:
                    caption += "["
                else:
                    caption += " "

                if self.flips[i] == -1 and i == self.x:
                    caption += "-"

                if i == self.x:
                    caption += "x"
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += "]"
                else:
                    caption += " "
            caption += "]"

            self.ax.set_title(caption)
            self.ax.axis("on")
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title("")
            self.fig.suptitle("")
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)


class ScatterPlot(object):
    """Plot array as scatter.

    Keyword Args:
        z: toggle current dimension as z dimension
        left/right: increment/decrement current dimension
        up/down: flip axis when current dimension is x or y
            otherwise increment/decrement slice at current dimension
        h: toggle hide all labels, titles and axes
        m: magnitude mode
        p: phase mode
        r: real mode
        i: imaginary mode
        l: log mode
    """

    def __init__(
        self,
        coord,
        data=None,
        z=None,
        hide_axes=False,
        mode="m",
        title="",
        save_basename="Figure",
        fps=10,
    ):
        import matplotlib.pyplot as plt

        self.coord = coord
        assert coord.shape[-1] == 2
        if data is None:
            self.data = np.ones(coord.shape[:-1])
        else:
            self.data = data

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)
        self.ax.set_facecolor("k")
        self.ax.axis("equal")

        for c, d in zip(coord.shape[:-1], self.data.shape[-coord.ndim + 1 :]):
            assert c == d

        self.ndim = self.data.ndim - self.coord.ndim + 1
        self.shape = self.data.shape[: self.ndim]

        self.slices = [s // 2 for s in self.shape]
        self.flips = [1] * self.ndim
        self.z = z % self.ndim if z is not None else None
        self.d = 0
        self.hide_axes = hide_axes
        self.title = title
        self.mode = mode
        self.axsc = None
        self.entering_slice = False
        self.save_basename = save_basename
        self.fps = fps
        self.vmin = None
        self.vmax = None

        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id
        )
        self.fig.canvas.mpl_connect("key_press_event", self.key_press)
        self.update_axes()
        self.update_data()
        self.fig.canvas.draw()
        plt.show()

    def key_press(self, event):
        if event.key == "up":
            if self.d != self.z:
                self.slices[self.d] = (self.slices[self.d] + 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == "down":
            if self.d != self.z:
                self.slices[self.d] = (self.slices[self.d] - 1) % self.shape[
                    self.d
                ]
            else:
                self.flips[self.d] *= -1

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == "left":
            self.d = (self.d - 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "right":
            self.d = (self.d + 1) % self.ndim

            self.update_axes()
            self.fig.canvas.draw()

        # elif event.key == 'z':
        #     if self.d == self.z:
        #         self.z = None
        #     else:
        #         self.z = self.d

        #     self.update_axes()
        #     self.update_data()
        #     self.fig.canvas.draw()

        elif event.key == "a":
            self.hide_axes = not self.hide_axes

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "f":
            self.fig.canvas.manager.full_screen_toggle()

        elif (
            event.key == "m"
            or event.key == "p"
            or event.key == "r"
            or event.key == "i"
            or event.key == "l"
        ):
            self.mode = event.key
            self.vmin = None
            self.vmax = None

            self.update_axes()
            self.update_data()
            self.fig.canvas.draw()

        elif event.key == "s":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.png"
            )
            self.fig.savefig(
                filename,
                transparent=True,
                format="png",
                bbox_inches="tight",
                pad_inches=0,
            )

        elif event.key == "g":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %I.%M.%S %p.gif"
            )
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_data()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    bbox_inches=bbox,
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vf",
                    "palettegen",
                    "{} palette.png".format(temp_basename),
                ]
            )
            subprocess.run(
                [
                    "ffmpeg",
                    "-f",
                    "image2",
                    "-s",
                    "{}x{}".format(
                        int(bbox.width * self.fig.dpi),
                        int(bbox.height * self.fig.dpi),
                    ),
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-i",
                    "{} palette.png".format(temp_basename),
                    "-lavfi",
                    "paletteuse",
                    filename,
                ]
            )

            os.remove("{} palette.png".format(temp_basename))
            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))

        elif event.key == "v":
            filename = self.save_basename + datetime.datetime.now().strftime(
                " %Y-%m-%d at %h.%M.%S %p.mov"
            )
            temp_basename = uuid.uuid4()

            bbox = self.fig.get_tightbbox(self.fig.canvas.get_renderer())
            for i in range(self.shape[self.d]):
                self.slices[self.d] = i

                self.update_axes()
                self.update_data()
                self.fig.canvas.draw()
                self.fig.savefig(
                    "{} {:05d}.png".format(temp_basename, i),
                    format="png",
                    bbox_inches=bbox,
                    pad_inches=0,
                )

            subprocess.run(
                [
                    "ffmpeg",
                    "-framerate",
                    str(self.fps),
                    "-i",
                    "{} %05d.png".format(temp_basename),
                    "-vcodec",
                    "png",
                    filename,
                ]
            )

            for i in range(self.shape[self.d]):
                os.remove("{} {:05d}.png".format(temp_basename, i))

        elif (
            event.key
            in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "backspace"]
            and self.d != self.z
        ):
            if self.entering_slice:
                if event.key == "backspace":
                    if self.entered_slice < 10:
                        self.entering_slice = False
                    else:
                        self.entered_slice //= 10
                else:
                    self.entered_slice = self.entered_slice * 10 + int(
                        event.key
                    )
            else:
                self.entering_slice = True
                self.entered_slice = int(event.key)

            self.update_axes()
            self.fig.canvas.draw()

        elif event.key == "enter" and self.entering_slice:
            self.entering_slice = False
            if self.entered_slice < self.shape[self.d]:
                self.slices[self.d] = self.entered_slice

                self.update_data()

            self.update_axes()
            self.fig.canvas.draw()

        else:
            return

    def update_data(self):
        idx = []
        for i in range(self.ndim):
            if i == self.z:
                idx.append(slice(None, None, self.flips[i]))
            else:
                idx.append(self.slices[i])

        idx = tuple(idx)
        if idx:
            datav = sp.to_device(self.data[idx])
        else:
            datav = sp.to_device(self.data)

        # if self.z is not None:
        #     datav_dims = [self.z] + datav_dims
        coordv = sp.to_device(self.coord)

        if self.mode == "m":
            datav = np.abs(datav)
        elif self.mode == "p":
            datav = np.angle(datav)
        elif self.mode == "r":
            datav = np.real(datav)
        elif self.mode == "i":
            datav = np.imag(datav)
        elif self.mode == "l":
            eps = 1e-31
            datav = np.log(np.abs(datav) + eps)

        datav = datav.ravel()
        if self.vmin is None:
            if datav.min() == datav.max():
                self.vmin = 0
            else:
                self.vmin = datav.min()

        if self.vmax is None:
            self.vmax = datav.max()

        if self.axsc is None:
            self.axsc = self.ax.scatter(
                coordv[..., 0].ravel(),
                coordv[..., 1].ravel(),
                c=datav,
                s=1,
                linewidths=0,
                cmap="gray",
                vmin=self.vmin,
                vmax=self.vmax,
            )

        else:
            self.axsc.set_offsets(coordv.T.reshape([-1, 2]))
            self.axsc.set_color(datav)

    def update_axes(self):
        if not self.hide_axes:
            caption = "["
            for i in range(self.ndim):
                if i == self.d:
                    caption += "["
                else:
                    caption += " "

                if self.flips[i] == -1 and i == self.z:
                    caption += "-"

                if i == self.z:
                    caption += "z"
                elif i == self.d and self.entering_slice:
                    caption += str(self.entered_slice) + "_"
                else:
                    caption += str(self.slices[i])

                if i == self.d:
                    caption += "]"
                else:
                    caption += " "
            caption += "]"

            self.ax.set_title(caption)
            self.fig.suptitle(self.title)
            self.ax.xaxis.set_visible(True)
            self.ax.yaxis.set_visible(True)
            self.ax.title.set_visible(True)
        else:
            self.ax.set_title("")
            self.fig.suptitle("")
            self.ax.xaxis.set_visible(False)
            self.ax.yaxis.set_visible(False)
            self.ax.title.set_visible(False)



# -*- coding: utf-8 -*-
"""Functions for simulations.

"""
import numpy as np

__all__ = ["shepp_logan"]


def shepp_logan(shape, dtype=complex):
    """Generates a Shepp Logan phantom with a given shape and dtype.

    Args:
        shape (tuple of ints): shape, can be of length 2 or 3.
        dtype (Dtype): data type.

    Returns:
        array.

    """
    return phantom(shape, sl_amps, sl_scales, sl_offsets, sl_angles, dtype)


sl_amps = [1, -0.8, -0.2, -0.2, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

sl_scales = [
    [0.6900, 0.920, 0.810],  # white big
    [0.6624, 0.874, 0.780],  # gray big
    [0.1100, 0.310, 0.220],  # right black
    [0.1600, 0.410, 0.280],  # left black
    [0.2100, 0.250, 0.410],  # gray center blob
    [0.0460, 0.046, 0.050],
    [0.0460, 0.046, 0.050],
    [0.0460, 0.046, 0.050],  # left small dot
    [0.0230, 0.023, 0.020],  # mid small dot
    [0.0230, 0.023, 0.020],
]

sl_offsets = [
    [0.0, 0.0, 0],
    [0.0, -0.0184, 0],
    [0.22, 0.0, 0],
    [-0.22, 0.0, 0],
    [0.0, 0.35, -0.15],
    [0.0, 0.1, 0.25],
    [0.0, -0.1, 0.25],
    [-0.08, -0.605, 0],
    [0.0, -0.606, 0],
    [0.06, -0.605, 0],
]

sl_angles = [
    [0, 0, 0],
    [0, 0, 0],
    [-18, 0, 10],
    [18, 0, 10],
    [0, 0, 0],
    [0, 0, 0],
    [0, 0, 0],
    [0, 0, 0],
    [0, 0, 0],
    [0, 0, 0],
]


def phantom(shape, amps, scales, offsets, angles, dtype):
    """
    Generate a cube of given shape using a list of ellipsoid
    parameters.
    """

    if len(shape) == 2:
        ndim = 2
        shape = (1, shape[-2], shape[-1])

    elif len(shape) == 3:
        ndim = 3

    else:
        raise ValueError("Incorrect dimension")

    out = np.zeros(shape, dtype=dtype)

    z, y, x = np.mgrid[
        -(shape[-3] // 2) : ((shape[-3] + 1) // 2),
        -(shape[-2] // 2) : ((shape[-2] + 1) // 2),
        -(shape[-1] // 2) : ((shape[-1] + 1) // 2),
    ]

    coords = np.stack(
        (
            x.ravel() / shape[-1] * 2,
            y.ravel() / shape[-2] * 2,
            z.ravel() / shape[-3] * 2,
        )
    )

    for amp, scale, offset, angle in zip(amps, scales, offsets, angles):
        ellipsoid(amp, scale, offset, angle, coords, out)

    if ndim == 2:
        return out[0, :, :]

    else:
        return out


def ellipsoid(amp, scale, offset, angle, coords, out):
    """
    Generate a cube containing an ellipsoid defined by its parameters.
    If out is given, fills the given cube instead of creating a new
    one.
    """
    R = rotation_matrix(angle)
    coords = (np.matmul(R, coords) - np.reshape(offset, (3, 1))) / np.reshape(
        scale, (3, 1)
    )

    r2 = np.sum(coords**2, axis=0).reshape(out.shape)

    out[r2 <= 1] += amp


def rotation_matrix(angle):
    cphi = np.cos(np.radians(angle[0]))
    sphi = np.sin(np.radians(angle[0]))
    ctheta = np.cos(np.radians(angle[1]))
    stheta = np.sin(np.radians(angle[1]))
    cpsi = np.cos(np.radians(angle[2]))
    spsi = np.sin(np.radians(angle[2]))
    alpha = [
        [
            cpsi * cphi - ctheta * sphi * spsi,
            cpsi * sphi + ctheta * cphi * spsi,
            spsi * stheta,
        ],
        [
            -spsi * cphi - ctheta * sphi * cpsi,
            -spsi * sphi + ctheta * cphi * cpsi,
            cpsi * stheta,
        ],
        [stheta * sphi, -stheta * cphi, ctheta],
    ]
    return np.array(alpha)



# -*- coding: utf-8 -*-
"""Functions for interoperability between sigpy and pytorch.

"""
import numpy as np

from sigpy import backend, config

__all__ = ["to_pytorch", "from_pytorch", "to_pytorch_function"]


def to_pytorch(array, requires_grad=True):  # pragma: no cover
    """Zero-copy conversion from numpy/cupy array to pytorch tensor.

    For complex array input, returns a tensor with shape + [2],
    where tensor[..., 0] and tensor[..., 1] represent the real
    and imaginary.

    Args:
        array (numpy/cupy array): input.
        requires_grad(bool): Set .requires_grad output tensor
    Returns:
        PyTorch tensor.

    """
    import torch
    from torch.utils.dlpack import from_dlpack

    device = backend.get_device(array)
    if not np.issubdtype(array.dtype, np.floating):
        with device:
            shape = array.shape
            array = array.view(dtype=array.real.dtype)
            array = array.reshape(shape + (2,))

    if device == backend.cpu_device:
        tensor = torch.from_numpy(array)
    else:
        tensor = from_dlpack(array.toDlpack())

    tensor.requires_grad = requires_grad
    return tensor.contiguous()


def from_pytorch(tensor, iscomplex=False):  # pragma: no cover
    """Zero-copy conversion from pytorch tensor to numpy/cupy array.

    If iscomplex, then tensor must have the last dimension as 2,
    and the output will be viewed as a complex valued array.

    Args:
        tensor (PyTorch tensor): input.
        iscomplex (bool): whether input represents complex valued tensor.

    Returns:
        Numpy/cupy array.

    """
    from torch.utils.dlpack import to_dlpack

    device = tensor.device
    if device.type == "cpu":
        output = tensor.detach().contiguous().numpy()
    else:
        if config.cupy_enabled:
            import cupy as cp

            output = cp.fromDlpack(to_dlpack(tensor.contiguous()))
        else:
            raise TypeError(
                "CuPy not installed, "
                "but trying to convert GPU PyTorch Tensor."
            )

    if iscomplex:
        if output.shape[-1] != 2:
            raise ValueError(
                "shape[-1] must be 2 when iscomplex is "
                "specified, but got {}".format(output.shape)
            )

        with backend.get_device(output):
            if output.dtype == np.float32:
                output = output.view(np.complex64)
            elif output.dtype == np.float64:
                output = output.view(np.complex128)

            output = output.reshape(output.shape[:-1])

    return output


def to_pytorch_function(
    linop, input_iscomplex=False, output_iscomplex=False
):  # pragma: no cover
    """Convert SigPy Linop to PyTorch Function.

    The returned function can be treated as a native
    pytorch function performing the linop operator.
    The function can be backpropagated, applied on GPU arrays,
    and has minimal overhead as the underlying arrays
    are shared without copying.
    For complex valued input/output, the appropriate options
    should be set when calling the function.

    Args:
        linop (Linop): linear operator to be converted.
        input_iscomplex (bool): whether the PyTorch input
            represents complex tensor.
        output_iscomplex (bool): whether the PyTorch output
            represents complex tensor.

    Returns:
        torch.autograd.Function: equivalent PyTorch Function.

    """
    import torch

    class LinopFunction(torch.autograd.Function):
        @staticmethod
        def forward(ctx, input):
            return to_pytorch(
                linop(from_pytorch(input, iscomplex=input_iscomplex))
            )

        @staticmethod
        def backward(ctx, grad_output):
            return to_pytorch(
                linop.H(from_pytorch(grad_output, iscomplex=output_iscomplex))
            )

    return LinopFunction



import unittest

import numpy as np
import numpy.testing as npt

import sigpy as sp
from sigpy.mri import dcf, samp

if __name__ == "__main__":
    unittest.main()


class TestApp(unittest.TestCase):
    def shepp_logan_setup(self):
        img_shape = [16, 16]
        coord_shape = [int(16 * np.pi), 16, 2]

        img = sp.shepp_logan(img_shape)
        coord = samp.radial(coord_shape, img_shape)
        ksp = sp.nufft(img, coord)
        return img, coord, ksp

    def test_shepp_logan_dcf(self):
        img, coord, ksp = self.shepp_logan_setup()
        pm_dcf = dcf.pipe_menon_dcf(coord, show_pbar=False)
        img_dcf = sp.nufft_adjoint(ksp * pm_dcf, coord, oshape=img.shape)
        img_dcf /= np.abs(img_dcf).max()
        npt.assert_allclose(img, img_dcf, atol=1, rtol=1e-1)



import unittest

import numpy as np

import sigpy as sp
import sigpy.mri.rf as rf
from sigpy.mri.rf import linop

if __name__ == "__main__":
    unittest.main()


def check_linop_adjoint(A, dtype=np.float32, device=sp.cpu_device):
    device = sp.Device(device)
    x = sp.randn(A.ishape, dtype=dtype, device=device)
    y = sp.randn(A.oshape, dtype=dtype, device=device)

    xp = device.xp
    with device:
        lhs = xp.vdot(A * x, y)
        rhs = xp.vdot(x, A.H * y)

        xp.testing.assert_allclose(lhs, rhs, atol=1e-5, rtol=1e-5)


class TestLinop(unittest.TestCase):
    def test_spatial_explicit_model(self):
        dim = 3
        img_shape = [dim, dim, dim]
        mps_shape = [8, dim, dim, dim]

        dt = 4e-6

        k = sp.mri.spiral(
            fov=dim / 2,
            N=dim,
            f_sampling=1,
            R=1,
            ninterleaves=1,
            alpha=1,
            gm=0.03,
            sm=200,
        )
        k = rf.stack_of(k, 3, 0.1)

        mps = sp.randn(mps_shape, dtype=np.complex64)

        A = linop.PtxSpatialExplicit(mps, k, dt, img_shape)

        check_linop_adjoint(A, dtype=np.complex64)



# -*- coding: utf-8 -*-
"""MRI gradient and excitation trajectory design
"""

import math

import numpy as np
from scipy import integrate, interpolate

from sigpy import util

__all__ = [
    "min_trap_grad",
    "trap_grad",
    "spiral_varden",
    "spiral_arch",
    "epi",
    "rosette",
    "spokes_grad",
    "stack_of",
    "traj_array_to_complex",
    "traj_complex_to_array",
    "min_time_gradient",
]


def min_trap_grad(area, gmax, dgdt, dt):
    r"""Minimal duration trapezoidal gradient designer. Design for target area
    under the flat portion (for non-ramp-sampled pulses)

    Args:
        area (float): pulse area in (g*sec)/cm
        gmax (float): maximum gradient in g/cm
        dgdt (float): max slew rate in g/cm/sec
        dt (float): sample time in sec

    Returns:
        2-element tuple containing

        - **trap** (*array*): gradient waveform in g/cm.
        - **ramppts** (*int*): number of points in ramps.

    """

    if np.abs(area) > 0:
        # we get the solution for plateau amp by setting derivative of
        # duration as a function of amplitude to zero and solving
        a = np.sqrt(dgdt * area / 2)

        # finish design with discretization
        # make a flat portion of magnitude a and enough area for the swath
        pts = np.floor(area / a / dt)
        flat = np.ones((1, int(pts)))
        flat = flat / np.sum(flat) * area / dt
        if np.max(flat) > gmax:
            flat = np.ones((1, int(np.ceil(area / gmax / dt))))
            flat = flat / np.sum(flat) * area / dt

        # make attack and decay ramps
        ramppts = int(np.ceil(np.max(flat) / dgdt / dt))
        ramp_up = (
            np.linspace(0, ramppts, num=ramppts + 1) / ramppts * np.max(flat)
        )
        ramp_dn = (
            np.linspace(ramppts, 0, num=ramppts + 1) / ramppts * np.max(flat)
        )

        trap = np.concatenate((ramp_up, np.squeeze(flat), ramp_dn))

    else:
        # negative-area trap requested?
        trap, ramppts = 0, 0

    return np.expand_dims(trap, axis=0), ramppts


def trap_grad(area, gmax, dgdt, dt, *args):
    r"""General trapezoidal gradient designer for total target area
    (for rewinders)

    Args:
        area (float): pulse area in (g*sec)/cm
        gmax (float): maximum gradient in g/cm
        dgdt (float): max slew rate in g/cm/sec
        dt (float): sample time in sec

    Returns:
        2-element tuple containing

        - **trap** (*array*): gradient waveform in g/cm.
        - **ramppts** (*int*): number of points in ramps.

    """

    if len(args) < 5:
        # in case we are making a rewinder
        rampsamp = 1

    if np.abs(area) > 0:
        if rampsamp:
            ramppts = int(np.ceil(gmax / dgdt / dt))
            triareamax = ramppts * dt * gmax

            if triareamax > np.abs(area):
                # triangle pulse
                newgmax = np.sqrt(np.abs(area) * dgdt)
                ramppts = int(np.ceil(newgmax / dgdt / dt))
                ramp_up = np.linspace(0, ramppts, num=ramppts + 1) / ramppts
                ramp_dn = np.linspace(ramppts, 0, num=ramppts + 1) / ramppts
                pulse = np.concatenate((ramp_up, ramp_dn))
            else:
                # trapezoid pulse
                nflat = int(np.ceil((area - triareamax) / gmax / dt / 2) * 2)
                ramp_up = np.linspace(0, ramppts, num=ramppts + 1) / ramppts
                ramp_dn = np.linspace(ramppts, 0, num=ramppts + 1) / ramppts
                pulse = np.concatenate((ramp_up, np.ones(nflat), ramp_dn))

            trap = pulse * (area / (sum(pulse) * dt))

        else:
            # make a flat portion of magnitude gmax
            # and enough area for the entire swath
            flat = np.ones(1, np.ceil(area / gmax / dt))
            flat = flat / sum(flat) * area / dt
            flat_top = np.max(flat)

            # make attack and decay ramps
            ramppts = int(np.ceil(np.max(flat) / dgdt / dt))
            ramp_up = (
                np.linspace(0, ramppts, num=ramppts + 1) / ramppts * flat_top
            )
            ramp_dn = (
                np.linspace(ramppts, 0, num=ramppts + 1) / ramppts * flat_top
            )
            trap = np.concatenate((ramp_up, flat, ramp_dn))

    else:
        trap, ramppts = 0, 0

    return np.expand_dims(trap, axis=0), ramppts


def spiral_varden(
    fov, res, gts, gslew, gamp, densamp, dentrans, nl, rewinder=False
):
    r"""Variable density spiral designer. Produces trajectory, gradients,
    and slew rate. Gradient units returned are in g/cm, g/cm/s

    Args:
        fov (float): imaging field of view (cm).
        res (float): imaging isotropic resolution (cm).
        gts (float): gradient sample time in sec.
        gslew (float): max slew rate in g/cm/s.
        gamp (float): max gradient amplitude in g/cm.
        densamp (float):  duration of full density sampling (# of samples).
        dentrans (float): duration of transition from higher to lower
            (should be >= densamp/2).
        nl (float): degree of undersampling outer region.
        rewinder (Boolean): if True, include rewinder. If false, exclude.

    Returns:
        tuple: (g, k, t, s, dens) tuple containing

        - **g** - (array): gradient waveform [g/cm]
        - **k** - (array): exact k-space corresponding to gradient g.
        - **time** - (array):  sampled time
        - **s** - (array): slew rate [g/cm/s]
        - **dens** - (array): undersampling factor at each time point.

    References:
        Code and algorithm based on spiralgradlx6 from
        Doug Noll, U. of Michigan BME
    """
    fsgcm = gamp  # fullscale g/cm
    risetime = gamp / gslew * 10000  # us
    ts = gts  # sampling time
    gts = gts  # gradient sampling time
    N = np.floor(fov / res)
    targetk = N / 2
    A = 32766  # output scaling of waveform (fullscale)

    max_dec_ratio = 32
    gam = 4257.0
    S = (gts / 1e-6) * A / risetime
    dr = ts / gts
    OMF = 2.0 * np.pi * fov / (1 / (gam * fsgcm * gts))
    OM = 2.0 * np.pi / nl * fov / (1 / (gam * fsgcm * gts))
    distance = 1.0 / (fov * gam * fsgcm * gts / A)

    ac = A
    loop = 1
    absk = 0
    dec_ratio = 1
    s0 = gslew * 100
    ggx, ggy = [], []
    dens = []
    kx, ky = [], []

    while loop > 0:
        loop = 0
        om = OM / dec_ratio
        omf = OMF / dec_ratio
        s = S / dec_ratio
        g0 = 0
        gx = g0
        gy = 0
        absg = np.abs(g0)
        oldkx = 0
        oldky = 0
        tkx = gx
        tky = gy
        kxt = tkx
        kyt = tky
        thetan_1 = 0
        taun = 0
        n = 0
        den1 = 0

        while absk < targetk:
            realn = n / dec_ratio
            taun_1 = taun
            taun = np.abs(tkx + 1j * tky) / A
            tauhat = taun
            if realn > densamp:
                if den1 == 0:
                    den1 = 1

                if realn > densamp + dentrans:
                    if "scthat" not in locals():
                        scthat = 0
                    scoffset = scthat
                    denoffset = taun_1
                    scthat = scoffset + om * (tauhat - denoffset)
                    fractrans = 1

                else:
                    scoffset = scthat
                    denoffset = taun_1
                    fractrans = (realn - densamp) / dentrans
                    fractrans = 1 - ((fractrans - 1) * (fractrans - 1))
                    scthat = omf + (om - omf) * fractrans
                    scthat *= tauhat - denoffset
                    scthat += scoffset

            else:
                fractrans = 0
                scthat = omf * tauhat

            theta = np.arctan2(scthat, 1.0) + scthat

            if absg < ac:
                deltheta = theta - thetan_1
                B = 1.0 / (1.0 + np.tan(deltheta) * np.tan(deltheta))
                gtilde = absg
                t1 = s * s
                t2 = gtilde * gtilde * (1 - B)

                if t2 > t1:
                    dec_ratio = dec_ratio * 2.0

                    if dec_ratio > max_dec_ratio:
                        print("k-space calculation failed.\n")
                        return

                    loop = 1
                    break

                t3 = np.sqrt(t1 - t2)
                absg = np.sqrt(B) * gtilde + t3

                if absg > ac:
                    absg = ac

            tgx = absg * np.cos(theta)
            tgy = absg * np.sin(theta)
            tkx += tgx
            tky += tgy
            thetan_1 = theta

            if np.remainder(n, dec_ratio) == 0:
                m = int(np.round(n / dec_ratio))
                gx = np.round((tkx - oldkx) / dec_ratio)
                gx = gx - np.remainder(gx, 2)
                gy = np.round((tky - oldky) / dec_ratio)
                gy = gy - np.remainder(gy, 2)
                if m > len(ggx) - 1:
                    ggx.append(gx)
                    ggy.append(gy)
                else:
                    ggx[m] = gx
                    ggy[m] = gy
                kxt = kxt + gx
                kyt = kyt + gy
                oldkx = tkx
                oldky = tky

                if np.remainder(m, dr) == 0:
                    m = int(m / dr)
                    absk = np.abs(kxt + 1j * kyt) / distance

                    if m > len(dens) - 1:
                        dens.append(omf / (omf + (om - omf) * fractrans))
                        if absk > targetk:
                            break
                        kx.append(kxt / distance)
                        ky.append(kyt / distance)
                    else:
                        dens[m] = omf / (omf + (om - omf) * fractrans)
                        if absk > targetk:
                            break
                        kx[m] = kxt / distance
                        ky[m] = kyt / distance

            n += 1

    g = []
    for i in range(len(ggx)):
        g.append(complex(ggx[i], ggy[i]) / A * fsgcm)
    dt = gts * 1000
    delk = 1 / 4.258 / fov  # (g ms)/cm

    # ramp down
    l2 = len(g) - 1
    rsteps = int(np.ceil(np.abs(g[l2]) / (s0 * 0.99) / gts))
    ind3 = l2 + np.linspace(1, rsteps, num=rsteps)
    c = g[l2] * np.linspace(rsteps, 0, num=rsteps) / rsteps
    g.extend(c)
    dens.extend([0] * len(ind3))

    # rewinder
    if rewinder:
        rewx, ramppts = np.squeeze(
            trap_grad(abs(np.real(sum(g))) * gts, gamp, gslew * 50, gts)
        )
        rewy, ramppts = np.squeeze(
            trap_grad(abs(np.imag(sum(g))) * gts, gamp, gslew * 50, gts)
        )

        # append rewinder gradient
        if len(rewx) > len(rewy):
            r = -np.sign(np.real(sum(g))) * rewx
            p = np.sign(np.imag(sum(g)))
            p *= 1j * np.abs(np.imag(sum(g))) / np.real(sum(g)) * rewx
            r -= p
        else:
            p = -np.sign(np.real(sum(g)))
            p *= np.abs(np.real(sum(g)) / np.imag(sum(g))) * rewy
            r = p - 1j * np.sign(np.imag(sum(g))) * rewy

        g = np.concatenate((g, r))

    # change from (real, imag) notation to (Nt, 2) notation
    gtemp = np.zeros((len(g), 2))
    gtemp[:, 0] = np.real(g)
    gtemp[:, 1] = np.imag(g)
    g = gtemp

    # calculate trajectory, slew rate factor from designed gradient
    k = np.cumsum(g, axis=0) * dt / delk / fov  # trajectory
    t = np.linspace(0, len(g), num=len(g) + 1)  # time vector
    s = np.diff(g, axis=0) / (gts * 1000)  # slew rate factor

    return g, k, t, s, dens


def spiral_arch(fov, res, gts, gslew, gamp):
    r"""Analytic Archimedean spiral designer. Produces trajectory, gradients,
    and slew rate. Gradient returned has units mT/m.

    Args:
        fov (float): imaging field of view in m.
        res (float): resolution, in m.
        gts (float): sample time in s.
        gslew (float): max slew rate in mT/m/ms.
        gamp (float): max gradient amplitude in mT/m.

    Returns:
        tuple: (g, k, t, s) tuple containing

        - **g** - (array): gradient waveform [mT/m]
        - **k** - (array): exact k-space corresponding to gradient g.
        - **time** - (array):  sampled time
        - **s** - (array): slew rate [mT/m/ms]

    References:
        Glover, G. H.(1999).
        Simple Analytic Spiral K-Space Algorithm.
        Magnetic resonance in medicine, 42, 412-415.

        Bernstein, M.A.; King, K.F.; amd Zhou, X.J. (2004).
        Handbook of MRI Pulse Sequences. Elsevier.
    """

    gam = 267.522 * 1e6 / 1000  # rad/s/mT
    gambar = gam / 2 / np.pi  # Hz/mT
    N = int(fov / res)  # effective matrix size
    lam = 1 / (2 * np.pi * fov)
    beta = gambar * gslew / lam

    kmax = N / (2 * fov)
    a_2 = (9 * beta / 4) ** (1 / 3)  # rad ** (1/3) / s ** (2/3)
    lamb = 5
    theta_max = kmax / lam
    ts = (3 * gam * gamp / (4 * np.pi * lam * a_2**2)) ** 3
    theta_s = 0.5 * beta * ts**2
    theta_s /= lamb + beta / (2 * a_2) * ts ** (4 / 3)
    t_g = np.pi * lam * (theta_max**2 - theta_s**2) / (gam * gamp)
    n_s = int(np.round(ts / gts))
    n_g = int(np.round(t_g / gts))

    if theta_max > theta_s:
        print(" Spiral trajectory is slewrate limited or amplitude limited")

        tacq = ts + t_g

        t_s = np.linspace(0, ts, n_s)
        t_g = np.linspace(ts + gts, tacq, n_g)

        theta_1 = beta / 2 * t_s**2
        theta_1 /= lamb + beta / (2 * a_2) * t_s ** (4 / 3)
        theta_2 = theta_s**2 + gam / (np.pi * lam) * gamp * (t_g - ts)
        theta_2 = np.sqrt(theta_2)

        k1 = lam * theta_1 * (np.cos(theta_1) + 1j * np.sin(theta_1))
        k2 = lam * theta_2 * (np.cos(theta_2) + 1j * np.sin(theta_2))
        k = np.concatenate((k1, k2), axis=0)

    else:
        tacq = 2 * np.pi * fov / 3 * np.sqrt(np.pi / (gam * gslew * res**3))
        n_t = int(np.round(tacq / gts))
        t_s = np.linspace(0, tacq, n_t)
        theta_1 = beta / 2 * t_s**2
        theta_1 /= lamb + beta / (2 * a_2) * t_s ** (4 / 3)

        k = lam * theta_1 * (np.cos(theta_1) + 1j * np.sin(theta_1))

    # end of trajectory calculation; prepare outputs
    g = np.diff(k, 1, axis=0) / (gts * gambar)  # gradient
    g = np.pad(g, (0, 1), "constant")
    s = np.diff(g, 1, axis=0) / (gts * 1000)  # slew rate factor
    s = np.pad(s, (0, 1), "constant")

    # change from (real, imag) notation to (Nt, 2) notation
    k = traj_complex_to_array(k)
    g = traj_complex_to_array(g)
    s = traj_complex_to_array(s)

    t = np.linspace(0, len(g), num=len(g) + 1)  # time vector

    return g, k, t, s


def epi(fov, n, etl, dt, gamp, gslew, offset=0, dirx=-1, diry=1):
    r"""Basic EPI single-shot trajectory designer.

    Args:
        fov (float): imaging field of view in cm.
        n (int): # of pixels (square). N = etl*nl, where etl = echo-train-len
            and nl = # leaves (shots). nl default 1.
        etl (int): echo train length.
        dt (float): sample time in s.
        gamp (float): max gradient amplitude in mT/m.
        gslew (float): max slew rate in mT/m/ms.
        offset (int): used for multi-shot EPI goes from 0 to #shots-1
        dirx (int): x direction of EPI -1 left to right, 1 right to left
        diry (int): y direction of EPI -1 bottom-top, 1 top-bottom

    Returns:
        tuple: (g, k, t, s) tuple containing

        - **g** - (array): gradient waveform [mT/m]
        - **k** - (array): exact k-space corresponding to gradient g.
        - **time** - (array):  sampled time
        - **s** - (array): slew rate [mT/m/ms]


    References:
        From Antonis Matakos' contrib to Jeff Fessler's IRT.
    """
    s = gslew * dt * 1000

    scaley = 20

    # make the various gradient waveforms
    gamma = 4.2575  # kHz/Gauss
    g = (1 / (1000 * dt)) / (gamma * fov)  # Gauss/cm
    if g > gamp:
        g = gamp
        print("max g reduced to {}".format(g))

    # readout trapezoid
    gxro = g * np.ones((1, n))  # plateau of readout trapezoid
    areapd = np.sum(gxro) * dt

    ramp = np.expand_dims(np.linspace(s, g, int(g / s)), axis=0)
    gxro = np.concatenate(
        (np.expand_dims(np.array([0]), axis=1), ramp, gxro, np.fliplr(ramp)),
        axis=1,
    )

    # x prewinder. make sure res_kpre is even. Handle even N by changing prew.
    if n % 2 == 0:
        area = (np.sum(gxro) - dirx * g) * dt
    else:
        area = np.sum(gxro) * dt
    gxprew = dirx * trap_grad(area / 2, gamp, gslew * 1000, dt)[0]

    gxprew = np.concatenate(
        (np.zeros((1, (gxprew.size + ramp.size) % 2)), gxprew), axis=1
    )

    # partial dephaser (one cycle of phase across each voxel)
    gxpd = -trap_grad(areapd / 2, gamp, gslew * 1000, dt)[0]
    gxpd = np.concatenate((np.zeros((1, gxpd.size % 2)), gxpd), axis=1)

    # phase-encode trapezoids before/after gx
    # handle even N by changing prewinder
    if n % 2 == 0:
        areayprew = areapd / 2 - offset * g * dt
    else:
        areayprew = (areapd - g * dt) / 2 - offset * g * dt

    gyprew = diry * trap_grad(areayprew, gamp, gslew / scaley * 1000, dt)[0]
    gyprew = np.concatenate((np.zeros((1, gyprew.size % 2)), gyprew), axis=1)

    lx = gxpd.size
    ly = gyprew.size
    if lx > ly:
        gyprew = np.concatenate((gyprew, np.zeros((1, lx - ly))), axis=1)
    else:
        gxpd = np.concatenate((gxpd, np.zeros((1, ly - lx))), axis=1)

    # gy readout gradient elements
    # changed readout patterns to create interleaved EPIs
    areagyblip = areapd / etl
    gyblip = trap_grad(areagyblip, gamp, gslew / scaley * 1000, dt)[0]
    gyro = np.concatenate(
        (np.zeros((1, gxro.size - gyblip.size)), gyblip), axis=1
    )
    gyro2 = np.expand_dims(np.array([0]), axis=1)

    # put together gx and gy

    gxro = -dirx * gxro
    gx = gxprew

    gyro = -diry * gyro
    gyro2 = -diry * gyro2
    gy = np.expand_dims(np.array([0]), axis=1)
    lx = gx.size
    ly = gy.size
    if lx > ly:
        gy = np.concatenate((gy, np.zeros((1, lx - ly))), axis=1)
    else:
        gx = np.concatenate((gx, np.zeros((1, ly - lx))), axis=1)

    gy = np.concatenate((gy, np.zeros((1, int(gyblip.size / 2)))), axis=1)

    for ee in range(1, etl):
        flip = (-1) ** (ee + 1)
        gx = np.concatenate((gx, flip * gxro), axis=1)
        gy = np.concatenate((gy, gyro), axis=1)

    if etl == 1:
        ee = 1
    else:
        ee += 1

    # concatenate with added 0 to limit max s
    gx = np.concatenate(
        (gx, (-(1 ** (ee + 1)) * gxro), np.expand_dims(np.array([0]), axis=1)),
        axis=1,
    )
    gy = np.concatenate((gy, np.zeros((1, gx.size - gy.size))), axis=1)

    # add rephasers at end of gx and gy readout
    areagx = np.sum(gx) * dt
    gxrep = trap_grad(-areagx, gamp, gslew * 1000, dt)[0]
    gx = np.concatenate((gx, gxrep), axis=1)

    areagy = np.sum(gy) * dt  # units = G/cm*s
    gyrep = trap_grad(-areagy, gamp, gslew / scaley * 1000, dt)[0]
    gy = np.concatenate((gy, gyrep), axis=1)

    # make sure length of gx and gy are same, and even
    lx = gx.size
    ly = gy.size
    if lx > ly:
        gy = np.concatenate((gy, np.zeros((1, lx - ly))), axis=1)
    else:
        gx = np.concatenate((gx, np.zeros((1, ly - lx))), axis=1)

    gx = np.concatenate((gx, np.zeros((1, gx.size % 2))), axis=1)
    gy = np.concatenate((gy, np.zeros((1, gy.size % 2))), axis=1)
    g = np.concatenate((gx, gy), axis=0)

    sx = np.diff(gx, axis=1) / (dt * 1000)
    sy = np.diff(gy, axis=1) / (dt * 1000)
    s = np.concatenate((sx, sy), axis=0)

    kx = np.cumsum(gx, axis=1) * gamma * dt * 1000
    ky = np.cumsum(gy, axis=1) * gamma * dt * 1000
    k = np.concatenate((kx, ky), axis=0)

    t = np.linspace(0, kx.size, kx.size) * dt

    return g, k, t, s


def rosette(kmax, w1, w2, dt, dur, gamp=None, gslew=None):
    r"""Basic rosette trajectory designer.

    Args:
        kmax (float): 1/m.
        w1 (float): rotational frequency (Hz).
        w2 (float): center sampling frequency (Hz).
        dt (float): sample time (s).
        dur (float): total duration (s).
        gamp (float): max gradient amplitude (mT/m).
        gslew (float): max slew rate (mT/m/ms).

    Returns:
        tuple: (g, k, t, s) tuple containing

        - **g** - (array): gradient waveform [mT/m]
        - **k** - (array): exact k-space corresponding to gradient g.
        - **time** - (array):  sampled time
        - **s** - (array): slew rate [mT/m/ms]

    References:
        D. C. Noll, 'Multi-shot rosette trajectories for spectrally selective
        MR imaging.' IEEE Trans. Med Imaging 16, 372-377 (1997).
    """

    # check if violates gradient or slew rate constraints
    gam = 267.522 * 1e6 / 1000  # rad/s/mT
    gambar = gam / 2 / np.pi  # Hz/mT
    if gamp is not None:
        if (1 / gambar) * kmax * w1 > gamp:
            print("gmax exceeded, decrease rosette kmax or w1")
            return
    if gslew is not None:
        if (1 / gambar) * kmax * (w1**2 + w2**2) / 1000 > gslew:
            print("smax exceeded, dcrease rosette kmax, w1, or w2")
            return
    t = np.linspace(0, dur, dur / dt)
    k = kmax * np.sin(w1 * t) * np.exp(1j * w2 * t)

    # end of trajectory calculation; prepare outputs
    g = np.diff(k, 1, axis=0) / (dt * gambar)  # gradient
    g = np.pad(g, (0, 1), "constant")
    s = np.diff(g, 1, axis=0) / (dt * 1000)  # slew rate factor
    s = np.pad(s, (0, 1), "constant")

    # change from (real, imag) notation to (Nt, 2) notation
    k = traj_complex_to_array(k)
    g = traj_complex_to_array(g)
    s = traj_complex_to_array(s)

    t = np.linspace(0, len(g), num=len(g) + 1)  # time vector

    return g, k, t, s


def spokes_grad(k, tbw, sl_thick, gmax, dgdtmax, gts):
    r"""Spokes gradient designer. Given some chosen spoke locations k, return
    the gradients required to move between those spoke locations.

    Args:
        k (array): spokes locations, [Nspokes, 2]
        tbw (int): time bandwidth product.
        sl_thick (float): slice thickness (mm).
        gmax (float): max gradient amplitude (g/cm).
        dgdtmax (float): max gradient slew (g/cm/s).
        gts (float): hardware sampling dwell time (s).

    Returns:
        g (array): gz, gy, and gz waveforms  in g/cm [3, Nt]

    References:
           Grissom, W., Khalighi, M., Sacolick, L., Rutt, B. & Vogel, M (2012).
           Small-tip-angle spokes pulse design using interleaved greedy and
           local optimization methods. Magnetic Resonance in Medicine, 68(5),
           1553-62.

    """
    n_spokes = k.shape[0]

    area = tbw / (sl_thick / 10) / 4257  # thick * kwid = twb, kwid = gam*area
    [subgz, nramp] = min_trap_grad(area, gmax, dgdtmax, gts)

    # calc gradient, add extra 0 location at end for return to (0, 0)
    gxarea = np.diff(np.concatenate((k[:, 0], np.zeros(1)))) / 4257
    gyarea = np.diff(np.concatenate((k[:, 1], np.zeros(1)))) / 4257

    gx, gy, gz = [], [], []
    gz_sign = -1
    for ii in range(n_spokes):
        gz_sign *= -1
        gz.extend(np.squeeze(gz_sign * subgz).tolist())  # alt sign of gz

        gx.extend([0] * np.size(subgz))  # zeros for gz duration
        if np.absolute(gxarea[ii]) > 0:
            [gblip, _] = trap_grad(abs(gxarea[ii]), gmax, dgdtmax, gts)
            gxblip = int(np.sign(gxarea[ii])) * gblip
            gx = gx[: len(gx) - len(gxblip.T)]
            gx.extend(np.squeeze(gxblip).tolist())

        gy.extend([0] * np.size(subgz))
        if np.absolute(gyarea[ii]) > 0:
            [gblip, _] = trap_grad(abs(gyarea[ii]), gmax, dgdtmax, gts)
            gyblip = int(np.sign(gyarea[ii])) * gblip
            gy = gy[: len(gy) - len(gyblip.T)]
            gy.extend(np.squeeze(gyblip).tolist())

    [gref, _] = trap_grad(gts * np.sum(subgz) / 2, gmax, dgdtmax, gts)
    gzref = -gref
    gz.extend(np.squeeze(gzref).tolist())
    gx.extend([0] * np.size(gzref))
    gy.extend([0] * np.size(gzref))

    # combine gradient waveforms
    gx = np.array(gx)
    g = np.vstack((np.array(gx), np.array(gy), np.array(gz)))

    return g


def stack_of(k, num, zres):
    r"""Function for creating a 3D stack of ____ trajectory from a 2D [Nt 2]
    trajectory.

    Args:
        k (array): 2D array in [2 x Nt]. Will be bottom of stack.
        num (int): number of layers of stack.
        zres (float): spacing between stacks in cm.
    """

    z = np.linspace(-num * zres / 2, num * zres / 2, num)
    kout = np.zeros((k.shape[0] * num, 3))

    # we will be performing a complex rotation on our trajectory
    k = traj_array_to_complex(k)

    for ii in range(num):
        kr = k[0:] * np.exp(2 * np.pi * 1j * ii / num)
        z_coord = np.expand_dims(np.ones(len(kr)) * z[ii], axis=1)
        krz = np.concatenate((traj_complex_to_array(kr), z_coord), axis=1)

        kout[ii * len(krz) : (ii + 1) * len(krz), :] = krz

    return kout


def traj_complex_to_array(k):
    r"""Function to convert complex convention trajectory to [Nt 2] trajectory

    Args:
        k (complex array): Nt vector
    """
    kout = np.zeros((len(k), 2))
    kout[:, 0], kout[:, 1] = np.real(k), np.imag(k)
    return kout


def traj_array_to_complex(k):
    r"""Function to convert [Nt 2] convention traj to complex convention

    Args:
        k (complex array): Nt vector
    """
    kout = k[:, 0] + 1j * k[:, 1]
    return kout


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def runge_kutta(
    ds: float, st: float, kvals: np.ndarray, smax=None, gamma=4.257
):
    r"""Runge-Kutta 4 for curve constrained

    Args:
        ds (float): spacing in arc length space
        st (float): output shape.
        kvals (array): 3 points of curve.
        smax (float): maximum slew
        gamma (float): gyromagnetic ratio

    Returns:
        float or None: step size dsdt or None
    """
    temp = gamma**2 * smax**2 - abs(kvals[0]) ** 2 * st**4
    if temp < 0.0:
        return None
    k1 = ds / st * math.sqrt(temp)

    temp = (
        gamma**2 * smax**2 - abs(kvals[1]) ** 2 * (st + ds * k1 / 2) ** 4
    )
    if temp < 0.0:
        return None
    k2 = ds / (st + ds * k1 / 2) * math.sqrt(temp)

    temp = (
        gamma**2 * smax**2 - abs(kvals[1]) ** 2 * (st + ds * k2 / 2) ** 4
    )
    if temp < 0.0:
        return None
    k3 = ds / (st + ds * k2 / 2) * math.sqrt(temp)

    temp = gamma**2 * smax**2 - abs(kvals[2]) ** 2 * (st + ds * k3) ** 4
    if temp < 0.0:
        return None
    k4 = ds / (st + ds * k3) * math.sqrt(temp)

    return k1 / 6 + k2 / 3 + k3 / 3 + k4 / 6


#  Arc length code translated from matlab
#    (c) Michael Lustig 2005
#    modified 2006 and 2007
#    Rewritten in Python in 2020 by Kevin Johnson
def min_time_gradient(
    c: np.ndarray, g0=0, gfin=None, gmax=4, smax=15, dt=4e-3, gamma=4.257
):
    r"""
    Given a k-space trajectory c(n), gradient and slew constraints. This
    function will return a new parametrization that will meet these
    constraint while getting from one point to the other in minimum time.

    Args:
        c (array): Curve in k-space given in any parametrization [1/cm]
                        Nx3 real array
        g0 (float): Initial gradient amplitude (leave empty for g0 = 0)
        gfin (float): Gradient value at the end of the trajectory. If not
                        possible, the result would be the largest possible
                        ampltude. (Leave empty if you don't care to get
                        maximum gradient.)
        gmax (float): Maximum gradient [G/cm] (3.9 default)
        smax (float): Maximum slew [G/Cm/ms]  (14.5 default)
        dt (float): Sampling time interval [ms] (4e-3 default)
        gamma (float): Gyromagnetic ratio

    Returns:
        tuple: (g, k, s, t) tuple containing

        - **g** - (array): gradient waveform [G/cm]
        - **k** - (array): exact k-space corresponding to gradient g.
        - **s** - (array): slew rate [G/cm/ms]
        - **time** - (array):  sampled time

    References:
        Lustig M, Kim SJ, Pauly JM. A fast method for designing time-optimal
        gradient waveforms for arbitrary k-space trajectories. IEEE Trans Med
        Imaging. 2008;27(6):866-873. doi:10.1109/TMI.2008.922699
    """

    def sdotmax(
        cs: interpolate.CubicSpline, s: np.ndarray, gmax, smax, gamma=4.257
    ):
        # [sdot, k, ] = sdotMax(PP, p_of_s, s, gmax, smax)
        #
        # Given a k-space curve C (in [1/cm] units), maximum gradient amplitude
        # (in G/cm) and maximum slew-rate (in G/(cm*ms)).
        # This function calculates the upper bound for the time parametrization
        # sdot (which is a non scaled max gradient constaint) as a function
        # of s.
        #
        #   cs      --  spline polynomial
        #   p_of_s  --  parametrization vs arclength
        #   s       --  arclength parametrization (0->1)
        #   gmax    --  maximum gradient (G/cm)
        #   smax    --  maximum slew rate (G/ cm*ms)
        #
        #   returns the maximum sdot (1st derivative of s) as a function of
        #   arclength s
        #   Also, returns curvature as a function of s and length of curve (L)
        #
        #  (c) Michael Lustig 2005
        #  last modified 2006

        # Absolute value of 2nd derivative in curve space using cubic splines
        cs2 = cs.derivative(2)  # spline derivative
        cs2_highres = cs2(s)  # evaluated along arc length
        k = np.linalg.norm(cs2_highres, axis=1)  # magnitude

        # calc I constraint curve (maximum gradient)
        sdot1 = gamma * gmax * np.ones_like(s)

        # calc II constraint curve (curve curvature dependent)
        sdot2 = np.sqrt(gamma * smax / (k + np.finfo(float).eps))

        # calc total constraint
        sdot = np.minimum(sdot1, sdot2)

        return sdot, k

    # Curve in arbitrary paramater space, cubic spline
    num_p = c.shape[0]
    p = np.linspace(0, 1, num_p, endpoint=True)
    cp = interpolate.CubicSpline(p, c, axis=0)

    # Integrate absolute value to find length and s(arc) vs p(paramater)
    cp1_spline = cp.derivative()
    p_highres = np.linspace(0, 1, num_p * 10)
    cp1_highres = cp1_spline(p_highres)
    ds_p = np.linalg.norm(cp1_highres, axis=1)

    # s vs p to enable conversion
    s_of_p = integrate.cumtrapz(ds_p, p_highres, initial=0)
    curve_length = s_of_p[-1]

    # decide ds and compute st for the first point
    stt0 = gamma * smax  # always assumes first point is max slew
    st0 = stt0 * dt / 8  # start at 1/8 the gradient for accuracy close to g=0
    s0 = st0 * dt
    ds = s0 / 4.0  # smaller step size for numerical accuracy
    ns = int(curve_length / ds)

    if g0 is None:
        g0 = 0

    # s is arc length at high resolution
    s = np.linspace(0, curve_length, ns, endpoint=True)

    # Cubic spline at s positions (s of p)
    cp_highres = cp(p_highres)
    cs = interpolate.CubicSpline(s_of_p, cp_highres, axis=0)

    # compute constraints (forbidden line curve)
    phi, k = sdotmax(cs, s, gmax, smax)

    # extend for the Runge-Kutte method
    k = np.pad(k, (0, 3), "constant", constant_values=(0,))

    # Get the start
    sta = np.zeros_like(s)
    sta[0] = min(g0 * gamma + st0, gamma * gmax)

    # solve ODE forward
    for n in range(1, s.shape[0]):
        kpos = n
        dstds = runge_kutta(ds, sta[n - 1], k[kpos : kpos + 4], smax)

        if dstds is None:
            sta[n] = phi[n]
        else:
            tmpst = sta[n - 1] + dstds
            sta[n] = min(tmpst, phi[n])

    stb = 0 * s
    if gfin is None:
        stb[-1] = sta[-1]
    else:
        stb[-1] = min(max(gfin * gamma, st0), gamma * gmax)

    # solve ODE backwards
    for n in range(s.shape[0] - 2, 0, -1):
        kpos_end = n  # to 0
        kpos = kpos_end + 3
        dstds = runge_kutta(ds, stb[n + 1], k[kpos : (kpos - 3) : -1], smax)

        if dstds is None:
            stb[n] = phi[n - 1]
        else:
            tmpst = stb[n + 1] + dstds
            stb[n] = min(tmpst, phi[n - 1])

    # Fix last point which is indexed a bit off
    n = 0
    kpos_end = n
    kpos = kpos_end + 3
    dstds = runge_kutta(ds, stb[n + 1], k[kpos::-1], smax)
    if dstds is None:
        stb[n] = phi[n * 2 - 1]
    else:
        tmpst = stb[n + 1] + dstds
        stb[n] = min(tmpst, phi[n - 1])

    # take the minimum of the curves
    ds = s[1] - s[0]
    st_of_s = np.minimum(sta, stb)

    # compute time
    t_of_s = integrate.cumtrapz(1.0 / st_of_s, initial=0) * ds

    t = np.arange(0, t_of_s[-1] + np.finfo(float).eps, dt)

    t_of_s = interpolate.CubicSpline(t_of_s, s)
    s_of_t = t_of_s(t)
    c = np.squeeze(cs(s_of_t))

    g = np.diff(c, axis=0, append=np.zeros((1, 3))) / gamma / dt
    g[-1, :] = g[-2, :] + g[-2, :] - g[-3, :]

    k = integrate.cumtrapz(g, t, initial=0, axis=0) * gamma

    s = np.diff(g, axis=0) / dt

    return g, k, s, t



# -*- coding: utf-8 -*-
"""Wavelet transform functions.
"""
import numpy as np

from sigpy import backend, config, util

__all__ = ["fwt", "iwt"]

if config.pywt_enabled:
    import pywt  # noqa


def get_wavelet_shape(shape, wave_name="db4", axes=None, level=None):
    if not config.pywt_enabled:
        raise ImportError("pywt is not installed. Please install pywt.")

    zshape = [((i + 1) // 2) * 2 for i in shape]

    tmp = pywt.wavedecn(
        np.zeros(zshape), wave_name, mode="zero", axes=axes, level=level
    )
    tmp, coeff_slices = pywt.coeffs_to_array(tmp, axes=axes)
    oshape = tmp.shape

    return oshape, coeff_slices


def fwt(input, wave_name="db4", axes=None, level=None):
    """Forward wavelet transform.

    Args:
        input (array): Input array.
        axes (None or tuple of int): Axes to perform wavelet transform.
        wave_name (str): Wavelet name.
        level (None or int): Number of wavelet levels.
    """
    if not config.pywt_enabled:
        raise ImportError("pywt is not installed. Please install pywt.")

    device = backend.get_device(input)
    input = backend.to_device(input, backend.cpu_device)

    zshape = [((i + 1) // 2) * 2 for i in input.shape]
    zinput = util.resize(input, zshape)

    coeffs = pywt.wavedecn(
        zinput, wave_name, mode="zero", axes=axes, level=level
    )
    output, _ = pywt.coeffs_to_array(coeffs, axes=axes)

    output = backend.to_device(output, device)
    return output


def iwt(input, oshape, coeff_slices, wave_name="db4", axes=None, level=None):
    """Inverse wavelet transform.

    Args:
        input (array): Input array.
        oshape (tuple of ints): Output shape.
        coeff_slices (list of slice): Slices to split coefficients.
        axes (None or tuple of int): Axes to perform wavelet transform.
        wave_name (str): Wavelet name.
        level (None or int): Number of wavelet levels.
    """
    if not config.pywt_enabled:
        raise ImportError("pywt is not installed. Please install pywt.")

    device = backend.get_device(input)
    input = backend.to_device(input, backend.cpu_device)

    input = pywt.array_to_coeffs(input, coeff_slices, output_format="wavedecn")
    output = pywt.waverecn(input, wave_name, mode="zero", axes=axes)
    output = util.resize(output, oshape)

    output = backend.to_device(output, device)
    return output



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import fourier, util

if __name__ == "__main__":
    unittest.main()


class TestFourier(unittest.TestCase):
    def test_fft(self):
        input = np.array([0, 1, 0], dtype=complex)
        npt.assert_allclose(
            fourier.fft(input), np.ones(3) / 3**0.5, atol=1e-5
        )

        input = np.array([1, 1, 1], dtype=complex)
        npt.assert_allclose(fourier.fft(input), [0, 3**0.5, 0], atol=1e-5)

        input = util.randn([4, 5, 6])
        npt.assert_allclose(
            fourier.fft(input),
            np.fft.fftshift(
                np.fft.fftn(np.fft.ifftshift(input), norm="ortho")
            ),
            atol=1e-5,
        )

        input = np.array([0, 1, 0], dtype=complex)
        npt.assert_allclose(
            fourier.fft(input, oshape=[5]), np.ones(5) / 5**0.5, atol=1e-5
        )

    def test_fft_dtype(self):
        for dtype in [np.complex64, np.complex128]:
            input = np.array([0, 1, 0], dtype=dtype)
            output = fourier.fft(input)

            assert output.dtype == dtype

    def test_ifft(self):
        input = np.array([0, 1, 0], dtype=complex)
        npt.assert_allclose(
            fourier.ifft(input), np.ones(3) / 3**0.5, atol=1e-5
        )

        input = np.array([1, 1, 1], dtype=complex)
        npt.assert_allclose(fourier.ifft(input), [0, 3**0.5, 0], atol=1e-5)

        input = util.randn([4, 5, 6])
        npt.assert_allclose(
            fourier.ifft(input),
            np.fft.fftshift(
                np.fft.ifftn(np.fft.ifftshift(input), norm="ortho")
            ),
            atol=1e-5,
        )

        input = np.array([0, 1, 0], dtype=complex)
        npt.assert_allclose(
            fourier.ifft(input, oshape=[5]), np.ones(5) / 5**0.5, atol=1e-5
        )

    def test_nufft(self):
        # Check deltas
        input = np.array([0, 1, 0], complex)  # delta
        coord = np.array([[-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.array([1.0, 1.0, 1.0]) / (3**0.5),
            atol=0.01,
            rtol=0.01,
        )

        input = np.array([0, 0, 1, 0], complex)  # delta
        coord = np.array([[-2], [-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.array([1.0, 1.0, 1.0, 1.0]) / (4**0.5),
            atol=0.01,
            rtol=0.01,
        )

        input = np.array([0, 0, 1, 0, 0], complex)  # delta
        coord = np.array([[-2], [-1], [0], [1], [2]], float)

        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.ones(5) / (5**0.5),
            atol=0.01,
            rtol=0.01,
        )

        # Check shifted delta
        input = np.array([0, 0, 1], complex)  # shifted delta
        coord = np.array([[-1], [0], [1]], float)

        w = np.exp(-1j * 2.0 * np.pi / 3.0)
        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.array([w.conjugate(), 1.0, w]) / (3**0.5),
            atol=0.01,
            rtol=0.01,
        )

        input = np.array([0, 0, 0, 1], complex)  # delta
        coord = np.array([[-2], [-1], [0], [1]], float)

        w = np.exp(-1j * 2.0 * np.pi / 4.0)
        npt.assert_allclose(
            fourier.nufft(input, coord),
            np.array([w.conjugate() ** 2, w.conjugate(), 1.0, w]) / (4**0.5),
            atol=0.01,
            rtol=0.01,
        )

    def test_nufft_nd(self):
        for ndim in range(3):
            input = np.array([[0], [1], [0]], complex)
            coord = np.array([[-1, 0], [0, 0], [1, 0]], float)

            npt.assert_allclose(
                fourier.nufft(input, coord),
                np.array([1.0, 1.0, 1.0]) / 3**0.5,
                atol=0.01,
                rtol=0.01,
            )

    def test_nufft_adjoint(self):
        # Check deltas
        oshape = [3]
        input = np.array([1.0, 1.0, 1.0], dtype=complex) / 3**0.5
        coord = np.array([[-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([0, 1, 0]),
            atol=0.01,
            rtol=0.01,
        )

        oshape = [4]
        input = np.array([1.0, 1.0, 1.0, 1.0], dtype=complex) / 4**0.5
        coord = np.array([[-2], [-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([0, 0, 1, 0], complex),
            atol=0.01,
            rtol=0.01,
        )

        oshape = [5]
        input = np.ones(5, dtype=complex) / 5**0.5
        coord = np.array([[-2], [-1], [0], [1], [2]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([0, 0, 1, 0, 0], complex),
            atol=0.01,
            rtol=0.01,
        )

        # Check shifted delta
        oshape = [3]
        w = np.exp(-1j * 2.0 * np.pi / 3.0)
        input = np.array([w.conjugate(), 1.0, w]) / 3**0.5
        coord = np.array([[-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([0, 0, 1], complex),
            atol=0.01,
            rtol=0.01,
        )

        oshape = [4]
        w = np.exp(-1j * 2.0 * np.pi / 4.0)
        input = (
            np.array([w.conjugate() ** 2, w.conjugate(), 1.0, w]) / 4**0.5
        )
        coord = np.array([[-2], [-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([0, 0, 0, 1], complex),
            atol=0.01,
            rtol=0.01,
        )

    def test_nufft_adjoint_nd(self):
        oshape = [3, 1]

        input = np.array([1.0, 1.0, 1.0], dtype=complex) / 3**0.5
        coord = np.array([[-1, 0], [0, 0], [1, 0]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(input, coord, oshape),
            np.array([[0], [1], [0]], complex),
            atol=0.01,
            rtol=0.01,
        )

    def test_nufft_normal(self):
        # Check delta
        oshape = [3]
        input = np.array([0.0, 1.0, 0.0], dtype=complex)
        coord = np.array([[-1], [0], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(fourier.nufft(input, coord), coord, oshape),
            np.array([0, 1, 0]),
            atol=0.01,
            rtol=0.01,
        )

        # Check delta scale
        oshape = [3]
        input = np.array([0.0, 1.0, 0.0], dtype=complex)
        coord = np.array([[-1], [-0.5], [0], [0.5], [1]], float)

        npt.assert_allclose(
            fourier.nufft_adjoint(fourier.nufft(input, coord), coord, oshape)[
                len(input) // 2
            ],
            5 / 3,
            atol=0.01,
            rtol=0.01,
        )

    def test_nufft_ndft(self):
        n = 5
        w = np.exp(-1j * 2 * np.pi / n)
        coord = np.array([[-2], [0], [0.1]])
        w2 = w**-2
        w1 = w**0.1
        A = (
            np.array(
                [
                    [w2**-2, w2**-1, 1, w2, w2**2],
                    [1, 1, 1, 1, 1],
                    [w1**-2, w1**-1, 1, w1, w1**2],
                ]
            )
            / n**0.5
        )

        for i in range(n):
            input = np.zeros(n, complex)
            input[i] = 1
            npt.assert_allclose(
                A[:, i], fourier.nufft(input, coord), atol=0.01, rtol=0.01
            )

        n = 6
        w = np.exp(-1j * 2 * np.pi / n)
        coord = np.array([[-2], [0], [0.1]])
        w2 = w**-2
        w1 = w**0.1
        A = (
            np.array(
                [
                    [w2**-3, w2**-2, w2**-1, 1, w2, w2**2],
                    [1, 1, 1, 1, 1, 1],
                    [w1**-3, w1**-2, w1**-1, 1, w1, w1**2],
                ]
            )
            / n**0.5
        )

        for i in range(n):
            input = np.zeros(n, complex)
            input[i] = 1
            npt.assert_allclose(
                A[:, i], fourier.nufft(input, coord), atol=0.1, rtol=0.1
            )



# -*- coding: utf-8 -*-
"""Interpolation functions.
"""
import numpy as np

from sigpy import backend, config, util

__all__ = ["interpolate", "gridding"]


KERNELS = ["spline", "kaiser_bessel"]


def interpolate(input, coord, kernel="spline", width=2, param=1):
    r"""Interpolation from array to points specified by coordinates.

    Let :math:`x` be the input, :math:`y` be the output,
    :math:`c` be the coordinates, :math:`W` be the kernel width,
    and :math:`K` be the interpolation kernel, then the function computes,

    .. math ::
        y[j] = \sum_{i : \| i - c[j] \|_\infty \leq W / 2}
               K\left(\frac{i - c[j]}{W / 2}\right) x[i]

    There are two types of kernels: 'spline' and 'kaiser_bessel'.

    'spline' uses the cardinal B-spline functions as kernels.
    The order of the spline can be specified using param.
    For example, param=1 performs linear interpolation.
    Concretely, for param=0, :math:`K(x) = 1`,
    for param=1, :math:`K(x) = 1 - |x|`, and
    for param=2, :math:`K(x) = \frac{9}{8} (1 - |x|)^2`
    for :math:`|x| > \frac{1}{3}`
    and :math:`K(x) = \frac{3}{4} (1 - 3 x^2)` for :math:`|x| < \frac{1}{3}`.

    These function expressions are derived from the reference wikipedia
    page by shifting and scaling the range to -1 to 1.
    When the coordinates specifies a uniformly spaced grid,
    it is recommended to use the original scaling with width=param + 1
    so that the interpolation weights add up to one.

    'kaiser_bessel' uses the Kaiser-Bessel function as kernel.
    Concretely, :math:`K(x) = I_0(\beta \sqrt{1 - x^2})`,
    where :math:`I_0` is the modified Bessel function of the first kind.
    The beta parameter can be specified with param.
    The modified Bessel function of the first kind is approximated
    using the power series, following the reference.

    Args:
        input (array): Input array of shape.
        coord (array): Coordinate array of shape [..., ndim]
        width (float or tuple of floats): Interpolation kernel full-width.
        kernel (str): Interpolation kernel, {'spline', 'kaiser_bessel'}.
        param (float or tuple of floats): Kernel parameter.

    Returns:
        output (array): Output array.

    References:
        https://en.wikipedia.org/wiki/Spline_wavelet#Cardinal_B-splines_of_small_orders
        http://people.math.sfu.ca/~cbm/aands/page_378.htm
    """
    ndim = coord.shape[-1]

    batch_shape = input.shape[:-ndim]
    batch_size = util.prod(batch_shape)

    pts_shape = coord.shape[:-1]
    npts = util.prod(pts_shape)

    xp = backend.get_array_module(input)

    input = input.reshape([batch_size] + list(input.shape[-ndim:]))
    coord = coord.reshape([npts, ndim])
    output = xp.zeros([batch_size, npts], dtype=input.dtype)

    if np.isscalar(param):
        param = xp.array([param] * ndim, coord.dtype)
    else:
        param = xp.array(param, coord.dtype)

    if np.isscalar(width):
        width = xp.array([width] * ndim, coord.dtype)
    else:
        width = xp.array(width, coord.dtype)

    if xp == np:
        _interpolate[kernel][ndim - 1](output, input, coord, width, param)
    else:  # pragma: no cover
        _interpolate_cuda[kernel][ndim - 1](
            input, coord, width, param, output, size=npts
        )

    return output.reshape(batch_shape + pts_shape)


def gridding(input, coord, shape, kernel="spline", width=2, param=1):
    r"""Gridding of points specified by coordinates to array.

    Let :math:`y` be the input, :math:`x` be the output,
    :math:`c` be the coordinates, :math:`W` be the kernel width,
    and :math:`K` be the interpolation kernel, then the function computes,

    .. math ::
        x[i] = \sum_{j : \| i - c[j] \|_\infty \leq W / 2}
               K\left(\frac{i - c[j]}{W / 2}\right) y[j]

    There are two types of kernels: 'spline' and 'kaiser_bessel'.

    'spline' uses the cardinal B-spline functions as kernels.
    The order of the spline can be specified using param.
    For example, param=1 performs linear interpolation.
    Concretely, for param=0, :math:`K(x) = 1`,
    for param=1, :math:`K(x) = 1 - |x|`, and
    for param=2, :math:`K(x) = \frac{9}{8} (1 - |x|)^2`
    for :math:`|x| > \frac{1}{3}`
    and :math:`K(x) = \frac{3}{4} (1 - 3 x^2)` for :math:`|x| < \frac{1}{3}`.

    These function expressions are derived from the reference wikipedia
    page by shifting and scaling the range to -1 to 1.
    When the coordinates specifies a uniformly spaced grid,
    it is recommended to use the original scaling with width=param + 1
    so that the interpolation weights add up to one.

    'kaiser_bessel' uses the Kaiser-Bessel function as kernel.
    Concretely, :math:`K(x) = I_0(\beta \sqrt{1 - x^2})`,
    where :math:`I_0` is the modified Bessel function of the first kind.
    The beta parameter can be specified with param.
    The modified Bessel function of the first kind is approximated
    using the power series, following the reference.

    Args:
        input (array): Input array.
        coord (array): Coordinate array of shape [..., ndim]
        width (float or tuple of floats): Interpolation kernel full-width.
        kernel (str): Interpolation kernel, {"spline", "kaiser_bessel"}.
        param (float or tuple of floats): Kernel parameter.

    Returns:
        output (array): Output array.

    References:
        https://en.wikipedia.org/wiki/Spline_wavelet#Cardinal_B-splines_of_small_orders
        http://people.math.sfu.ca/~cbm/aands/page_378.htm
    """
    ndim = coord.shape[-1]

    batch_shape = shape[:-ndim]
    batch_size = util.prod(batch_shape)

    pts_shape = coord.shape[:-1]
    npts = util.prod(pts_shape)

    xp = backend.get_array_module(input)
    isreal = np.issubdtype(input.dtype, np.floating)

    input = input.reshape([batch_size, npts])
    coord = coord.reshape([npts, ndim])
    output = xp.zeros([batch_size] + list(shape[-ndim:]), dtype=input.dtype)

    if np.isscalar(param):
        param = xp.array([param] * ndim, coord.dtype)
    else:
        param = xp.array(param, coord.dtype)

    if np.isscalar(width):
        width = xp.array([width] * ndim, coord.dtype)
    else:
        width = xp.array(width, coord.dtype)

    if xp == np:
        _gridding[kernel][ndim - 1](output, input, coord, width, param)
    else:  # pragma: no cover
        if isreal:
            _gridding_cuda[kernel][ndim - 1](
                input, coord, width, param, output, size=npts
            )
        else:
            _gridding_cuda_complex[kernel][ndim - 1](
                input, coord, width, param, output, size=npts
            )

    return output.reshape(shape)


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _spline_kernel(x, order):
    if abs(x) > 1:
        return 0

    if order == 0:
        return 1
    elif order == 1:
        return 1 - abs(x)
    elif order == 2:
        if abs(x) > 1 / 3:
            return 9 / 8 * (1 - abs(x)) ** 2
        else:
            return 3 / 4 * (1 - 3 * x**2)


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _kaiser_bessel_kernel(x, beta):
    if abs(x) > 1:
        return 0

    x = beta * (1 - x**2) ** 0.5
    t = x / 3.75
    if x < 3.75:
        return (
            1
            + 3.5156229 * t**2
            + 3.0899424 * t**4
            + 1.2067492 * t**6
            + 0.2659732 * t**8
            + 0.0360768 * t**10
            + 0.0045813 * t**12
        )
    else:
        return (
            x**-0.5
            * np.exp(x)
            * (
                0.39894228
                + 0.01328592 * t**-1
                + 0.00225319 * t**-2
                - 0.00157565 * t**-3
                + 0.00916281 * t**-4
                - 0.02057706 * t**-5
                + 0.02635537 * t**-6
                - 0.01647633 * t**-7
                + 0.00392377 * t**-8
            )
        )


def _get_interpolate(kernel):
    if kernel == "spline":
        kernel = _spline_kernel
    elif kernel == "kaiser_bessel":
        kernel = _kaiser_bessel_kernel

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _interpolate1(output, input, coord, width, param):
        batch_size, nx = input.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx = coord[i, -1]

            x0 = np.ceil(kx - width[-1] / 2)
            x1 = np.floor(kx + width[-1] / 2)

            for x in range(x0, x1 + 1):
                w = kernel((x - kx) / (width[-1] / 2), param[-1])

                for b in range(batch_size):
                    output[b, i] += w * input[b, x % nx]

        return output

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _interpolate2(output, input, coord, width, param):
        batch_size, ny, nx = input.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx, ky = coord[i, -1], coord[i, -2]

            x0, y0 = (np.ceil(kx - width[-1] / 2), np.ceil(ky - width[-2] / 2))

            x1, y1 = (
                np.floor(kx + width[-1] / 2),
                np.floor(ky + width[-2] / 2),
            )

            for y in range(y0, y1 + 1):
                wy = kernel((y - ky) / (width[-2] / 2), param[-2])

                for x in range(x0, x1 + 1):
                    w = wy * kernel((x - kx) / (width[-1] / 2), param[-1])

                    for b in range(batch_size):
                        output[b, i] += w * input[b, y % ny, x % nx]

        return output

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _interpolate3(output, input, coord, width, param):
        batch_size, nz, ny, nx = input.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx, ky, kz = coord[i, -1], coord[i, -2], coord[i, -3]

            x0, y0, z0 = (
                np.ceil(kx - width[-1] / 2),
                np.ceil(ky - width[-2] / 2),
                np.ceil(kz - width[-3] / 2),
            )

            x1, y1, z1 = (
                np.floor(kx + width[-1] / 2),
                np.floor(ky + width[-2] / 2),
                np.floor(kz + width[-3] / 2),
            )

            for z in range(z0, z1 + 1):
                wz = kernel((z - kz) / (width[-3] / 2), param[-3])

                for y in range(y0, y1 + 1):
                    wy = wz * kernel((y - ky) / (width[-2] / 2), param[-2])

                    for x in range(x0, x1 + 1):
                        w = wy * kernel((x - kx) / (width[-1] / 2), param[-1])

                        for b in range(batch_size):
                            output[b, i] += (
                                w * input[b, z % nz, y % ny, x % nx]
                            )

        return output

    return _interpolate1, _interpolate2, _interpolate3


def _get_gridding(kernel):
    if kernel == "spline":
        kernel = _spline_kernel
    elif kernel == "kaiser_bessel":
        kernel = _kaiser_bessel_kernel

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _gridding1(output, input, coord, width, param):
        batch_size, nx = output.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx = coord[i, -1]

            x0 = np.ceil(kx - width[-1] / 2)
            x1 = np.floor(kx + width[-1] / 2)
            for x in range(x0, x1 + 1):
                w = kernel((x - kx) / (width[-1] / 2), param[-1])

                for b in range(batch_size):
                    output[b, x % nx] += w * input[b, i]

        return output

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _gridding2(output, input, coord, width, param):
        batch_size, ny, nx = output.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx, ky = coord[i, -1], coord[i, -2]

            x0, y0 = (np.ceil(kx - width[-1] / 2), np.ceil(ky - width[-2] / 2))

            x1, y1 = (
                np.floor(kx + width[-1] / 2),
                np.floor(ky + width[-2] / 2),
            )
            for y in range(y0, y1 + 1):
                wy = kernel((y - ky) / (width[-2] / 2), param[-2])
                for x in range(x0, x1 + 1):
                    w = wy * kernel((x - kx) / (width[-1] / 2), param[-1])

                    for b in range(batch_size):
                        output[b, y % ny, x % nx] += w * input[b, i]

        return output

    @util.fallback_jit(nopython=True)  # pragma: no cover
    def _gridding3(output, input, coord, width, param):
        batch_size, nz, ny, nx = output.shape
        npts = coord.shape[0]

        for i in range(npts):
            kx, ky, kz = coord[i, -1], coord[i, -2], coord[i, -3]

            x0, y0, z0 = (
                np.ceil(kx - width[-1] / 2),
                np.ceil(ky - width[-2] / 2),
                np.ceil(kz - width[-3] / 2),
            )

            x1, y1, z1 = (
                np.floor(kx + width[-1] / 2),
                np.floor(ky + width[-2] / 2),
                np.floor(kz + width[-3] / 2),
            )

            for z in range(z0, z1 + 1):
                wz = kernel((z - kz) / (width[-3] / 2), param[-3])

                for y in range(y0, y1 + 1):
                    wy = wz * kernel((y - ky) / (width[-2] / 2), param[-2])

                    for x in range(x0, x1 + 1):
                        w = wy * kernel((x - kx) / (width[-1] / 2), param[-1])

                        for b in range(batch_size):
                            output[b, z % nz, y % ny, x % nx] += (
                                w * input[b, i]
                            )

        return output

    return _gridding1, _gridding2, _gridding3


_interpolate = {}
_gridding = {}
for kernel in KERNELS:
    _interpolate[kernel] = _get_interpolate(kernel)
    _gridding[kernel] = _get_gridding(kernel)

if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _spline_kernel_cuda = """
    __device__ inline S kernel(S x, S order) {
        if (fabsf(x) > 1)
            return 0;

        if (order == 0)
            return 1;
        else if (order == 1)
            return 1 - fabsf(x);
        else if (fabsf(x) > 1 / 3)
            return 9 / 8 * (1 - fabsf(x)) * (1 - fabsf(x));
        else
            return 3 / 4 * (1 - 3 * x * x);
    }
    """

    _kaiser_bessel_kernel_cuda = """
    __device__ inline S kernel(S x, S beta) {
        if (fabsf(x) > 1)
            return 0;

        x = beta * sqrt(1 - x * x);
        S t = x / 3.75;
        S t2 = t * t;
        S t4 = t2 * t2;
        S t6 = t4 * t2;
        S t8 = t6 * t2;
        if (x < 3.75) {
            S t10 = t8 * t2;
            S t12 = t10 * t2;
            return 1 + 3.5156229 * t2 + 3.0899424 * t4 +
                1.2067492 * t6 + 0.2659732 * t8 +
                0.0360768 * t10 + 0.0045813 * t12;
        } else {
            S t3 = t * t2;
            S t5 = t3 * t2;
            S t7 = t5 * t2;

            return exp(x) / sqrt(x) * (
                0.39894228 + 0.01328592 / t +
                0.00225319 / t2 - 0.00157565 / t3 +
                0.00916281 / t4 - 0.02057706 / t5 +
                0.02635537 / t6 - 0.01647633 / t7 +
                0.00392377 / t8);
        }
    }
    """

    mod_cuda = """
    __device__ inline int mod(int x, int n) {
        return (x % n + n) % n;
    }
    """

    def _get_interpolate_cuda(kernel):
        if kernel == "spline":
            kernel = _spline_kernel_cuda
        elif kernel == "kaiser_bessel":
            kernel = _kaiser_bessel_kernel_cuda

        _interpolate1_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 1;
            const int batch_size = input.shape()[0];
            const int nx = input.shape()[1];

            const int coord_idx[] = {i, 0};
            const S kx = coord[coord_idx];
            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int x1 = floor(kx + width[ndim - 1] / 2.0);

            for (int x = x0; x < x1 + 1; x++) {
                const S w = kernel(
                    ((S) x - kx) / (width[ndim - 1] / 2.0), param[ndim - 1]);
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, mod(x, nx)};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, i};
                    output[output_idx] += v;
                }
            }
            """,
            name="interpolate1",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        _interpolate2_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 2;
            const int batch_size = input.shape()[0];
            const int ny = input.shape()[1];
            const int nx = input.shape()[2];

            const int coordx_idx[] = {i, 1};
            const S kx = coord[coordx_idx];
            const int coordy_idx[] = {i, 0};
            const S ky = coord[coordy_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);

            for (int y = y0; y < y1 + 1; y++) {
                const S wy = kernel(
                    ((S) y - ky) / (width[ndim - 2] / 2.0),
                    param[ndim - 2]);
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * kernel(
                        ((S) x - kx) / (width[ndim - 1] / 2.0),
                        param[ndim - 1]);
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, mod(y, ny), mod(x, nx)};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, i};
                        output[output_idx] += v;
                    }
                }
            }
            """,
            name="interpolate2",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        _interpolate3_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 3;
            const int batch_size = input.shape()[0];
            const int nz = input.shape()[1];
            const int ny = input.shape()[2];
            const int nx = input.shape()[3];

            const int coordz_idx[] = {i, 0};
            const S kz = coord[coordz_idx];
            const int coordy_idx[] = {i, 1};
            const S ky = coord[coordy_idx];
            const int coordx_idx[] = {i, 2};
            const S kx = coord[coordx_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);
            const int z0 = ceil(kz - width[ndim - 3] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);
            const int z1 = floor(kz + width[ndim - 3] / 2.0);

            for (int z = z0; z < z1 + 1; z++) {
                const S wz = kernel(
                    ((S) z - kz) / (width[ndim - 3] / 2.0),
                    param[ndim - 3]);
                for (int y = y0; y < y1 + 1; y++) {
                    const S wy = wz * kernel(
                        ((S) y - ky) / (width[ndim - 2] / 2.0),
                        param[ndim - 2]);
                    for (int x = x0; x < x1 + 1; x++) {
                        const S w = wy * kernel(
                            ((S) x - kx) / (width[ndim - 1] / 2.0),
                            param[ndim - 1]);
                        for (int b = 0; b < batch_size; b++) {
                            const int input_idx[] = {b, mod(z, nz), mod(y, ny),
                                mod(x, nx)};
                            const T v = (T) w * input[input_idx];
                            const int output_idx[] = {b, i};
                            output[output_idx] += v;
                        }
                    }
                }
            }
            """,
            name="interpolate3",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        return _interpolate1_cuda, _interpolate2_cuda, _interpolate3_cuda

    def _get_gridding_cuda(kernel):
        if kernel == "spline":
            kernel = _spline_kernel_cuda
        elif kernel == "kaiser_bessel":
            kernel = _kaiser_bessel_kernel_cuda

        _gridding1_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 1;
            const int batch_size = output.shape()[0];
            const int nx = output.shape()[1];

            const int coord_idx[] = {i, 0};
            const S kx = coord[coord_idx];
            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int x1 = floor(kx + width[ndim - 1] / 2.0);

            for (int x = x0; x < x1 + 1; x++) {
                const S w = kernel(
                    ((S) x - kx) / (width[ndim - 1] / 2.0), param[ndim - 1]);
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, i};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, mod(x, nx)};
                    atomicAdd(&output[output_idx], v);
                }
            }
            """,
            name="gridding1",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        _gridding2_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 2;
            const int batch_size = output.shape()[0];
            const int ny = output.shape()[1];
            const int nx = output.shape()[2];

            const int coordx_idx[] = {i, 1};
            const S kx = coord[coordx_idx];
            const int coordy_idx[] = {i, 0};
            const S ky = coord[coordy_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);

            for (int y = y0; y < y1 + 1; y++) {
                const S wy = kernel(
                    ((S) y - ky) / (width[ndim - 2] / 2.0),
                    param[ndim - 2]);
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * kernel(
                        ((S) x - kx) / (width[ndim - 1] / 2.0),
                        param[ndim - 1]);
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, i};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, mod(y, ny), mod(x, nx)};
                        atomicAdd(&output[output_idx], v);
                    }
                }
            }
            """,
            name="gridding2",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        _gridding3_cuda = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 3;
            const int batch_size = output.shape()[0];
            const int nz = output.shape()[1];
            const int ny = output.shape()[2];
            const int nx = output.shape()[3];

            const int coordz_idx[] = {i, 0};
            const S kz = coord[coordz_idx];
            const int coordy_idx[] = {i, 1};
            const S ky = coord[coordy_idx];
            const int coordx_idx[] = {i, 2};
            const S kx = coord[coordx_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);
            const int z0 = ceil(kz - width[ndim - 3] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);
            const int z1 = floor(kz + width[ndim - 3] / 2.0);

            for (int z = z0; z < z1 + 1; z++) {
                const S wz = kernel(
                    ((S) z - kz) / (width[ndim - 3] / 2.0),
                    param[ndim - 3]);
                for (int y = y0; y < y1 + 1; y++) {
                    const S wy = wz * kernel(
                        ((S) y - ky) / (width[ndim - 2] / 2.0),
                        param[ndim - 2]);
                    for (int x = x0; x < x1 + 1; x++) {
                        const S w = wy * kernel(
                            ((S) x - kx) / (width[ndim - 1] / 2.0),
                            param[ndim - 1]);
                        for (int b = 0; b < batch_size; b++) {
                            const int input_idx[] = {b, i};
                            const T v = (T) w * input[input_idx];
                            const int output_idx[] = {
                                b, mod(z, nz), mod(y, ny), mod(x, nx)};
                            atomicAdd(&output[output_idx], v);
                        }
                    }
                }
            }
            """,
            name="gridding3",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        return _gridding1_cuda, _gridding2_cuda, _gridding3_cuda

    def _get_gridding_cuda_complex(kernel):
        if kernel == "spline":
            kernel = _spline_kernel_cuda
        elif kernel == "kaiser_bessel":
            kernel = _kaiser_bessel_kernel_cuda

        _gridding1_cuda_complex = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 1;
            const int batch_size = output.shape()[0];
            const int nx = output.shape()[1];

            const int coord_idx[] = {i, 0};
            const S kx = coord[coord_idx];
            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int x1 = floor(kx + width[ndim - 1] / 2.0);

            for (int x = x0; x < x1 + 1; x++) {
                const S w = kernel(
                    ((S) x - kx) / (width[ndim - 1] / 2.0), param[ndim - 1]);
                for (int b = 0; b < batch_size; b++) {
                    const int input_idx[] = {b, i};
                    const T v = (T) w * input[input_idx];
                    const int output_idx[] = {b, mod(x, nx)};
                    atomicAdd(
                        reinterpret_cast<T::value_type*>(
                            &(output[output_idx])), v.real());
                    atomicAdd(
                        reinterpret_cast<T::value_type*>(
                            &(output[output_idx])) + 1, v.imag());
                }
            }
            """,
            name="gridding1_complex",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )
        _gridding2_cuda_complex = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 2;
            const int batch_size = output.shape()[0];
            const int ny = output.shape()[1];
            const int nx = output.shape()[2];

            const int coordx_idx[] = {i, 1};
            const S kx = coord[coordx_idx];
            const int coordy_idx[] = {i, 0};
            const S ky = coord[coordy_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);

            for (int y = y0; y < y1 + 1; y++) {
                const S wy = kernel(
                    ((S) y - ky) / (width[ndim - 2] / 2.0),
                    param[ndim - 2]);
                for (int x = x0; x < x1 + 1; x++) {
                    const S w = wy * kernel(
                        ((S) x - kx) / (width[ndim - 1] / 2.0),
                        param[ndim - 1]);
                    for (int b = 0; b < batch_size; b++) {
                        const int input_idx[] = {b, i};
                        const T v = (T) w * input[input_idx];
                        const int output_idx[] = {b, mod(y, ny), mod(x, nx)};
                        atomicAdd(reinterpret_cast<T::value_type*>(
                            &(output[output_idx])), v.real());
                        atomicAdd(reinterpret_cast<T::value_type*>(
                            &(output[output_idx])) + 1, v.imag());
                    }
                }
            }
            """,
            name="gridding2_complex",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        _gridding3_cuda_complex = cp.ElementwiseKernel(
            "raw T input, raw S coord, raw S width, raw S param",
            "raw T output",
            """
            const int ndim = 3;
            const int batch_size = output.shape()[0];
            const int nz = output.shape()[1];
            const int ny = output.shape()[2];
            const int nx = output.shape()[3];

            const int coordz_idx[] = {i, 0};
            const S kz = coord[coordz_idx];
            const int coordy_idx[] = {i, 1};
            const S ky = coord[coordy_idx];
            const int coordx_idx[] = {i, 2};
            const S kx = coord[coordx_idx];

            const int x0 = ceil(kx - width[ndim - 1] / 2.0);
            const int y0 = ceil(ky - width[ndim - 2] / 2.0);
            const int z0 = ceil(kz - width[ndim - 3] / 2.0);

            const int x1 = floor(kx + width[ndim - 1] / 2.0);
            const int y1 = floor(ky + width[ndim - 2] / 2.0);
            const int z1 = floor(kz + width[ndim - 3] / 2.0);

            for (int z = z0; z < z1 + 1; z++) {
                const S wz = kernel(
                    ((S) z - kz) / (width[ndim - 3] / 2.0),
                    param[ndim - 3]);
                for (int y = y0; y < y1 + 1; y++) {
                    const S wy = wz * kernel(
                         ((S) y - ky) / (width[ndim - 2] / 2.0),
                         param[ndim - 2]);
                    for (int x = x0; x < x1 + 1; x++) {
                        const S w = wy * kernel(
                            ((S) x - kx) / (width[ndim - 1] / 2.0),
                            param[ndim - 1]);
                        for (int b = 0; b < batch_size; b++) {
                            const int input_idx[] = {b, i};
                            const T v = (T) w * input[input_idx];
                            const int output_idx[] = {
                                b, mod(z, nz), mod(y, ny), mod(x, nx)};
                            atomicAdd(reinterpret_cast<T::value_type*>(
                                &(output[output_idx])), v.real());
                            atomicAdd(reinterpret_cast<T::value_type*>(
                                &(output[output_idx])) + 1, v.imag());
                        }
                    }
                }
            }
            """,
            name="gridding3_complex",
            preamble=kernel + mod_cuda,
            reduce_dims=False,
        )

        return (
            _gridding1_cuda_complex,
            _gridding2_cuda_complex,
            _gridding3_cuda_complex,
        )

    _interpolate_cuda = {}
    _gridding_cuda = {}
    _gridding_cuda_complex = {}
    for kernel in KERNELS:
        _interpolate_cuda[kernel] = _get_interpolate_cuda(kernel)
        _gridding_cuda[kernel] = _get_gridding_cuda(kernel)
        _gridding_cuda_complex[kernel] = _get_gridding_cuda_complex(kernel)



import unittest

import numpy as np
import numpy.testing as npt

import sigpy as sp
from sigpy.mri import app, sim

if __name__ == "__main__":
    unittest.main()


class TestApp(unittest.TestCase):
    def shepp_logan_setup(self):
        img_shape = [6, 6]
        mps_shape = [4, 6, 6]

        img = sp.shepp_logan(img_shape)
        mps = sim.birdcage_maps(mps_shape)

        mask = np.zeros(img_shape)
        mask[:, ::2] = 1

        ksp = mask * sp.fft(mps * img, axes=[-2, -1])
        return img, mps, ksp

    def test_shepp_logan_SenseRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0

        for solver in [
            "ConjugateGradient",
            "GradientMethod",
            "PrimalDualHybridGradient",
            "ADMM",
        ]:
            for coil_batch_size in [None, 1, 2, 3]:
                with self.subTest(solver=solver):
                    img_rec = app.SenseRecon(
                        ksp,
                        mps,
                        lamda,
                        solver=solver,
                        coil_batch_size=coil_batch_size,
                        show_pbar=False,
                    ).run()
                    npt.assert_allclose(img, img_rec, atol=1e-2, rtol=1e-2)

    if sp.config.mpi4py_enabled:

        def test_shepp_logan_SenseRecon_with_comm(self):
            img, mps, ksp = self.shepp_logan_setup()
            lamda = 0
            comm = sp.Communicator()
            ksp = ksp[comm.rank :: comm.size]
            mps = mps[comm.rank :: comm.size]

            for solver in [
                "ConjugateGradient",
                "GradientMethod",
                "PrimalDualHybridGradient",
                "ADMM",
            ]:
                with self.subTest(solver=solver):
                    img_rec = app.SenseRecon(
                        ksp,
                        mps,
                        lamda,
                        comm=comm,
                        solver=solver,
                        show_pbar=False,
                    ).run()
                    npt.assert_allclose(img, img_rec, atol=1e-2, rtol=1e-2)

    def test_shepp_logan_L1WaveletRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0

        for solver in ["GradientMethod", "PrimalDualHybridGradient", "ADMM"]:
            with self.subTest(solver=solver):
                img_rec = app.L1WaveletRecon(
                    ksp, mps, lamda, solver=solver, show_pbar=False
                ).run()
                npt.assert_allclose(img, img_rec, atol=1e-2, rtol=1e-2)

    def test_shepp_logan_TotalVariationRecon(self):
        img, mps, ksp = self.shepp_logan_setup()
        lamda = 0
        for solver in ["PrimalDualHybridGradient", "ADMM"]:
            with self.subTest(solver=solver):
                img_rec = app.TotalVariationRecon(
                    ksp,
                    mps,
                    lamda,
                    solver=solver,
                    max_iter=1000,
                    show_pbar=False,
                ).run()

                npt.assert_allclose(img, img_rec, atol=1e-2, rtol=1e-2)

    def test_ones_JsenseRecon(self):
        img_shape = [6, 6]
        mps_shape = [4, 6, 6]

        img = np.ones(img_shape, dtype=complex)
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps * img, axes=[-2, -1])

        _app = app.JsenseRecon(
            ksp, mps_ker_width=6, ksp_calib_width=6, show_pbar=False
        )
        mps_rec = _app.run()

        npt.assert_allclose(mps, mps_rec, atol=1e-2, rtol=1e-2)

    def test_espirit_maps(self):
        # 2D
        mps_shape = [8, 16, 16]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2])
        mps_rec = app.EspiritCalib(ksp, show_pbar=False).run()

        np.testing.assert_allclose(
            np.abs(mps)[:, 4:-4, 4:-4],
            np.abs(mps_rec[:, 4:-4, 4:-4]),
            rtol=1e-2,
            atol=1e-2,
        )

        # 3D
        mps_shape = [8, 16, 16, 16]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2, -3])
        mps_rec = app.EspiritCalib(ksp, show_pbar=False).run()

        np.testing.assert_allclose(
            np.abs(mps)[:, 4:-4, 4:-4, 4:-4],
            np.abs(mps_rec[:, 4:-4, 4:-4, 4:-4]),
            rtol=1e-2,
            atol=1e-2,
        )

    def test_espirit_maps_eig(self):
        # 2D
        mps_shape = [8, 16, 16]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2])
        mps_rec, eig_val = app.EspiritCalib(
            ksp, output_eigenvalue=True, show_pbar=False
        ).run()

        np.testing.assert_allclose(eig_val, 1, rtol=0.01, atol=0.01)

        # 3D
        mps_shape = [8, 16, 16, 16]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2, -3])
        mps_rec, eig_val = app.EspiritCalib(
            ksp, output_eigenvalue=True, show_pbar=False
        ).run()

        np.testing.assert_allclose(eig_val, 1, rtol=0.01, atol=0.01)



"""The module contains functions and classes for MRI reconstruction.

It provides convenient simulation and sampling functions,
such as the poisson-disc sampling function. It also
provides functions to compute preconditioners,
and density compensation factors.

"""
from sigpy.mri import app, dcf, linop, precond, samp, sim, util
from sigpy.mri.dcf import *  # noqa
from sigpy.mri.precond import *  # noqa
from sigpy.mri.samp import *  # noqa
from sigpy.mri.sim import *  # noqa
from sigpy.mri.util import *  # noqa

__all__ = ["app", "linop"]
__all__.extend(dcf.__all__)
__all__.extend(precond.__all__)
__all__.extend(samp.__all__)
__all__.extend(sim.__all__)
__all__.extend(util.__all__)



import math
import unittest

import numpy as np
import numpy.testing as npt

import sigpy.mri.rf as rf

if __name__ == "__main__":
    unittest.main()


class TestTrajGrad(unittest.TestCase):
    def test_min_gradient(self):
        t = np.linspace(0, 1, 1000)
        kx = np.sin(2.0 * math.pi * t)
        ky = np.cos(2.0 * math.pi * t)
        kz = t
        k = np.stack((kx, ky, kz), axis=-1)

        (g, k, s, t) = rf.min_time_gradient(
            k, 0.0, 0.0, gmax=4, smax=15, dt=4e-3, gamma=4.257
        )

        npt.assert_almost_equal(np.max(t), 0.916, decimal=4)

    def test_trap_grad(self):
        dt = 4e-6  # s
        area = 200 * dt
        dgdt = 18000  # g/cm/s
        gmax = 2  # g/cm

        trap, _ = rf.trap_grad(area, gmax, dgdt, dt)

        npt.assert_almost_equal(area, np.sum(trap) * dt, decimal=3)
        npt.assert_almost_equal(gmax, np.max(trap), decimal=1)

    def test_min_trap_grad(self):
        dt = 4e-6  # s
        area = 200 * dt
        dgdt = 18000  # g/cm/s
        gmax = 2  # g/cm

        trap, _ = rf.min_trap_grad(area, gmax, dgdt, dt)

        npt.assert_almost_equal(area, np.sum(trap) * dt, decimal=3)
        npt.assert_almost_equal(gmax, np.max(trap), decimal=1)



# -*- coding: utf-8 -*-
"""MRI RF utilities.
"""

import numpy as np

__all__ = ["dinf"]


def dinf(d1=0.01, d2=0.01):
    """Calculate D infinity for a linear phase filter.

    Args:
        d1 (float): passband ripple level in M0**-1.
        d2 (float): stopband ripple level in M0**-1.

    Returns:
        float: D infinity.

    References:
        Pauly J, Le Roux P, Nishimra D, Macovski A. Parameter relations for the
        Shinnar-Le Roux selective excitation pulse design algorithm.
        IEEE Tr Medical Imaging 1991; 10(1):53-65.

    """

    a1 = 5.309e-3
    a2 = 7.114e-2
    a3 = -4.761e-1
    a4 = -2.66e-3
    a5 = -5.941e-1
    a6 = -4.278e-1

    l10d1 = np.log10(d1)
    l10d2 = np.log10(d2)

    d = (a1 * l10d1 * l10d1 + a2 * l10d1 + a3) * l10d2 + (
        a4 * l10d1 * l10d1 + a5 * l10d1 + a6
    )

    return d



import unittest

import numpy as np

from sigpy import config, interp

if config.cupy_enabled:
    import cupy as cp

if __name__ == "__main__":
    unittest.main()


class TestInterp(unittest.TestCase):
    def test_interpolate_spline(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        batch = 2
        for xp in xps:
            for ndim in [1, 2, 3]:
                for dtype in [np.float32, np.complex64]:
                    with self.subTest(ndim=ndim, xp=xp, dtype=dtype):
                        shape = [3] + [1] * (ndim - 1)
                        coord = xp.array(
                            [
                                [0.1] + [0] * (ndim - 1),
                                [1.1] + [0] * (ndim - 1),
                                [2.1] + [0] * (ndim - 1),
                            ]
                        )

                        input = xp.array([[0, 1.0, 0]] * batch, dtype=dtype)
                        input = input.reshape([batch] + shape)
                        output = interp.interpolate(input, coord)
                        output_expected = xp.array([[0.1, 0.9, 0]] * batch)
                        xp.testing.assert_allclose(
                            output, output_expected, atol=1e-7
                        )

    def test_gridding_spline(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        batch = 2
        for xp in xps:
            for ndim in [1, 2, 3]:
                for dtype in [np.float32, np.complex64]:
                    with self.subTest(ndim=ndim, xp=xp, dtype=dtype):
                        shape = [3] + [1] * (ndim - 1)
                        coord = xp.array(
                            [
                                [0.1] + [0] * (ndim - 1),
                                [1.1] + [0] * (ndim - 1),
                                [2.1] + [0] * (ndim - 1),
                            ]
                        )

                        input = xp.array([[0, 1.0, 0]] * batch, dtype=dtype)
                        output = interp.gridding(input, coord, [batch] + shape)
                        output_expected = xp.array(
                            [[0, 0.9, 0.1]] * batch
                        ).reshape([batch] + shape)
                        xp.testing.assert_allclose(
                            output, output_expected, atol=1e-7
                        )

    if config.cupy_enabled:

        def test_interpolate_cpu_gpu(self):
            for ndim in [1, 2, 3]:
                for dtype in [np.float32, np.complex64]:
                    with self.subTest(ndim=ndim, dtype=dtype):
                        shape = [2, 20] + [1] * (ndim - 1)
                        coord = np.random.random([10, ndim])

                        input = np.random.random(shape).astype(dtype=dtype)
                        output_cpu = interp.interpolate(
                            input, coord, kernel="kaiser_bessel"
                        )
                        output_gpu = interp.interpolate(
                            cp.array(input),
                            cp.array(coord),
                            kernel="kaiser_bessel",
                        ).get()
                        np.testing.assert_allclose(
                            output_cpu, output_gpu, atol=1e-5
                        )

        def test_gridding_cpu_gpu(self):
            for ndim in [1, 2, 3]:
                for dtype in [np.float32, np.complex64]:
                    with self.subTest(ndim=ndim, dtype=dtype):
                        shape = [2, 20] + [1] * (ndim - 1)
                        coord = np.random.random([10, ndim])

                        input = np.random.random(
                            [2, 10] + [1] * (ndim - 1)
                        ).astype(dtype=dtype)
                        output_cpu = interp.gridding(
                            input, coord, shape, kernel="kaiser_bessel"
                        )
                        output_gpu = interp.gridding(
                            cp.array(input),
                            cp.array(coord),
                            shape,
                            kernel="kaiser_bessel",
                        ).get()
                        np.testing.assert_allclose(
                            output_cpu, output_gpu, atol=1e-5
                        )



# -*- coding: utf-8 -*-
"""MRI simulation functions.
"""
import numpy as np

__all__ = ["birdcage_maps"]


def birdcage_maps(shape, r=1.5, nzz=8, dtype=complex):
    """Simulates birdcage coil sensitivies.

    Args:
        shape (tuple of ints): sensitivity maps shape,
            can be of length 3, and 4.
        r (float): relative radius of birdcage.
        nzz (int): number of coils per ring.
        dtype (Dtype): data type.

    Returns:
        array.
    """

    if len(shape) == 3:
        nc, ny, nx = shape
        c, y, x = np.mgrid[:nc, :ny, :nx]

        coilx = r * np.cos(c * (2 * np.pi / nc))
        coily = r * np.sin(c * (2 * np.pi / nc))
        coil_phs = -c * (2 * np.pi / nc)

        x_co = (x - nx / 2.0) / (nx / 2.0) - coilx
        y_co = (y - ny / 2.0) / (ny / 2.0) - coily
        rr = np.sqrt(x_co**2 + y_co**2)
        phi = np.arctan2(x_co, -y_co) + coil_phs
        out = (1.0 / rr) * np.exp(1j * phi)

    elif len(shape) == 4:
        nc, nz, ny, nx = shape
        c, z, y, x = np.mgrid[:nc, :nz, :ny, :nx]

        coilx = r * np.cos(c * (2 * np.pi / nzz))
        coily = r * np.sin(c * (2 * np.pi / nzz))
        coilz = np.floor(c / nzz) - 0.5 * (np.ceil(nc / nzz) - 1)
        coil_phs = -(c + np.floor(c / nzz)) * (2 * np.pi / nzz)

        x_co = (x - nx / 2.0) / (nx / 2.0) - coilx
        y_co = (y - ny / 2.0) / (ny / 2.0) - coily
        z_co = (z - nz / 2.0) / (nz / 2.0) - coilz
        rr = (x_co**2 + y_co**2 + z_co**2) ** 0.5
        phi = np.arctan2(x_co, -y_co) + coil_phs
        out = (1 / rr) * np.exp(1j * phi)
    else:
        raise ValueError("Can only generate shape with length 3 or 4")

    rss = sum(abs(out) ** 2, 0) ** 0.5
    out /= rss

    return out.astype(dtype)



import unittest

import numpy as np
import numpy.testing as npt

import sigpy.mri.rf as rf

if __name__ == "__main__":
    unittest.main()


class TestSim(unittest.TestCase):
    def test_abrm(self):
        #  also provides testing of SLR excitation. Check ex profile sim.
        tb = 8
        N = 128
        d1 = 0.01
        d2 = 0.01
        ptype = "ex"
        ftype = "ls"

        pulse = rf.slr.dzrf(N, tb, ptype, ftype, d1, d2, False)
        [a, b] = rf.sim.abrm(pulse, np.arange(-2 * tb, 2 * tb, 0.01), True)
        Mxy = 2 * np.multiply(np.conj(a), b)

        pts = np.array(
            [
                Mxy[int(len(Mxy) / 2 - len(Mxy) / 3)],
                Mxy[int(len(Mxy) / 2)],
                Mxy[int(len(Mxy) / 2 + len(Mxy) / 3)],
            ]
        )

        npt.assert_almost_equal(abs(pts), np.array([0, 1, 0]), decimal=2)
